from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.system_status import SystemStatus
from pytpp.tools.helpers.date_converter import from_date_string

class _SystemStatus(API):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/SystemStatus')
        self.Upgrade = self._Upgrade(api_obj=api_obj)
        self.Version = self._Version(api_obj=api_obj)

    def get(self):
        class _Response(APIResponse):
            def __init__(self, response):
                super().__init__(response=response)

            @property
            @api_response_property()
            def engines(self):
                return [SystemStatus.SystemStatus(status) for status in self._from_json()]

        return _Response(response=self._get())

    class _Upgrade:
        def __init__(self, api_obj):
            self.Engine = self._Engine(api_obj=api_obj)
            self.Engines = self._Engines(api_obj=api_obj)
            self.History = self._History(api_obj=api_obj)
            self.Status = self._Status(api_obj=api_obj)
            self.Summary = self._Summary(api_obj=api_obj)

        class _Engine(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/SystemStatus/Upgrade/Engine')

            def get(self, guid: str, engine_id: str, upgrade_id: str = None):
                params = {
                    'guid': guid,
                    'Id': engine_id,
                    'UpgradeId': upgrade_id
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def engine(self):
                        return SystemStatus.Engine(self._from_json(key='Engine'))

                    @property
                    @api_response_property()
                    def status(self) -> str:
                        return self._from_json(key='Status')

                    @property
                    @api_response_property()
                    def upgrade_start_time(self):
                        return from_date_string(self._from_json(key='UpgradeStartTime'))

                    @property
                    @api_response_property()
                    def upgrade_stop_time(self):
                        return from_date_string(self._from_json(key='UpgradeStopTime'))

                    @property
                    @api_response_property()
                    def tasks_completed(self):
                        return [SystemStatus.Task(task) for task in self._from_json(key='TasksCompleted')]

                    @property
                    @api_response_property()
                    def tasks_pending(self):
                        return [SystemStatus.Task(task) for task in self._from_json(key='TasksPending')]

                    @property
                    @api_response_property()
                    def tasks_running(self):
                        return [SystemStatus.Task(task) for task in self._from_json(key='TasksRunning')]

                return _Response(response=self._get(params=params))

        class _Engines(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/SystemStatus/Upgrade/Engines')

            def get(self, upgrade_id: int = None):
                params = {
                    'UpgradeId': upgrade_id
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def engines(self):
                        return [SystemStatus.UpgradeStatus(engine) for engine in self._from_json(key='Engines')]

                return _Response(response=self._get(params=params))

        class _History(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/SystemStatus/Upgrade/History')

            def get(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def upgrade_history(self):
                        return [SystemStatus.UpgradeInfo(info) for info in self._from_json(key='UpgradeHistory')]

                return _Response(response=self._get())

        class _Status(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/SystemStatus/Upgrade/Status')

            def get(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def upgrade_in_progress(self) -> bool:
                        return self._from_json(key='UpgradeInProgress')

                return _Response(response=self._get())

        class _Summary(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/SystemStatus/Upgrade/Summary')

            def get(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def upgrade_summary(self):
                        return SystemStatus.UpgradeSummary(self._from_json(key='UpgradeSummary'))

                return _Response(response=self._get())

    class _Version(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/SystemStatus/Version')

        def get(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def version(self) -> str:
                    return self._from_json('Version')

            return _Response(response=self._get())



from pytpp.attributes._helper import IterableMeta, Attribute


class ZoneBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Zone Base"
	address_range = Attribute('Address Range')
	window_end = Attribute('Window End')
	window_start = Attribute('Window Start')
	zone_contact = Attribute('Zone Contact')
	zone_description = Attribute('Zone Description')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class UserPreferenceAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "User Preference"
	console_timeout = Attribute('Console Timeout')
	user_preferences = Attribute('User Preferences')



class PermissionParameter:
    is_associate_allowed = 'IsAssociateAllowed'
    is_create_allowed = 'IsCreateAllowed'
    is_delete_allowed = 'IsDeleteAllowed'
    is_manage_permissions_allowed = 'IsManagePermissionsAllowed'
    is_policy_write_allowed = 'IsPolicyWriteAllowed'
    is_private_key_read_allowed = 'IsPrivateKeyReadAllowed'
    is_private_key_write_allowed = 'IsPrivateKeyWriteAllowed'
    is_read_allowed = 'IsReadAllowed'
    is_rename_allowed = 'IsRenameAllowed'
    is_revoke_allowed = 'IsRevokeAllowed'
    is_view_allowed = 'IsViewAllowed'
    is_write_allowed = 'IsWriteAllowed'


class ProviderTypes:
    local = 'local'
    ad = 'AD'
    ldap = 'LDAP'



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class LicenseAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "License"
	license_key = Attribute('License Key')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.connection_base import ConnectionBaseAttributes
from pytpp.attributes.driver_base import DriverBaseAttributes
from pytpp.attributes.validation_base import ValidationBaseAttributes


class ApplicationBaseAttributes(ConnectionBaseAttributes, DriverBaseAttributes, ValidationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Application Base"
	adaptable_workflow_approvers = Attribute('Adaptable Workflow Approvers', min_version='18.3')
	adaptable_workflow_reference_id = Attribute('Adaptable Workflow Reference ID', min_version='18.3')
	adaptable_workflow_stage = Attribute('Adaptable Workflow Stage', min_version='18.3')
	agent_validate_now = Attribute('Agent Validate Now', min_version='18.2')
	approver = Attribute('Approver')
	certificate = Attribute('Certificate')
	certificate_file = Attribute('Certificate File', min_version='16.2')
	certificate_installed = Attribute('Certificate Installed')
	created_by = Attribute('Created By', min_version='15.1')
	discovered_by_dn = Attribute('Discovered By DN', min_version='19.1')
	discovered_on = Attribute('Discovered On')
	file_owner_group = Attribute('File Owner: Group', min_version='16.3')
	file_owner_user = Attribute('File Owner: User', min_version='16.3')
	file_permissions_enabled = Attribute('File Permissions Enabled', min_version='16.3')
	file_permissions_group = Attribute('File Permissions: Group', min_version='16.3')
	file_permissions_user = Attribute('File Permissions: User', min_version='16.3')
	grouping_id = Attribute('Grouping Id', min_version='15.3')
	in_error = Attribute('In Error')
	in_process = Attribute('In Process', min_version='15.3')
	key_encryption_algorithm = Attribute('Key Encryption Algorithm')
	key_store_vault_id = Attribute('Key Store Vault Id', min_version='15.3')
	last_pushed_by = Attribute('Last Pushed By', min_version='16.2')
	last_pushed_on = Attribute('Last Pushed On', min_version='16.2')
	private_key_password_credential = Attribute('Private Key Password Credential')
	remote_one_to_many_generation = Attribute('Remote One To Many Generation', min_version='19.4')
	restart_application = Attribute('Restart Application')
	stage = Attribute('Stage')
	status = Attribute('Status')
	ticket_dn = Attribute('Ticket DN')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.agent_driver_base import AgentDriverBaseAttributes
from pytpp.attributes.agent_ssh_base import AgentSSHBaseAttributes
from pytpp.attributes.schedule_base import ScheduleBaseAttributes


class AgentSSHDriverAttributes(AgentDriverBaseAttributes, AgentSSHBaseAttributes, ScheduleBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Agent SSH Driver"



import base64
import hashlib
from dataclasses import dataclass
from typing import List, Union
from datetime import datetime
from pytpp.tools.vtypes import Config, Identity
from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import InvalidResultCode
from pytpp.properties.secret_store import KeyNames, Namespaces, VaultTypes
from pytpp.attributes.workflow import WorkflowAttributes
from pytpp.attributes.adaptable_workflow import AdaptableWorkflowAttributes
from pytpp.attributes.workflow_ticket import WorkflowTicketAttributes


class _WorkflowBase(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)

    def delete(self, workflow: 'Union[Config.Object, str]'):
        """
        Deletes a workflow.

        Args:
            workflow: :ref:`config_object` or :ref:`dn` of the workflow object.
        """
        workflow_dn = self._get_dn(workflow)
        self._secret_store_delete(object_dn=workflow_dn)
        self._config_delete(object_dn=workflow_dn)

    def get(self, workflow_dn: str, raise_error_if_not_exists: bool = True):
        """
        Returns the config object of the folder DN.

        Args:
            workflow_dn: :ref:`dn` of the workflow.
            raise_error_if_not_exists: Raise an exception if the object :ref:`dn` does not exist.

        Returns:
            :ref:`config_object` of the workflow object.
        """
        return self._get_config_object(
            object_dn=workflow_dn,
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def _create(self, name: str, parent_folder: 'Union[Config.Object, str]', is_adaptable: bool, stage: int, injection_command: str = None,
                application_class_name: str = None, approvers: str = None, reason_code: int = None, attributes: dict = None,
                get_if_already_exists: bool = True):    
        workflow = self._config_create(
            name=name,
            parent_folder_dn=self._get_dn(parent_folder),
            config_class=WorkflowAttributes.__config_class__ if not is_adaptable else AdaptableWorkflowAttributes.__config_class__,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )

        if is_adaptable:
            driver_str = '037Venafi.Drivers.AdaptableWFApplication'
        else:
            driver_str = '028Venafi.Drivers.WFApplication'
        if application_class_name:
            application_class_name = str(application_class_name)
            stage_str = f'B&{len(str(stage)):03d}{stage}D-{len(application_class_name):03d}{application_class_name}'
        else:
            stage_str = f'B-{len(str(stage)):03d}{stage}'

        injection_command_str = f'B&{len(injection_command):03d}{injection_command}' if injection_command else ''
        approvers_str = f'C-{"0" * (3 - len(str(len(approvers))))}{str(len(approvers))}{approvers}'
        reason_code_str = f'{"0" * (3 - len(str(reason_code)))}{reason_code}'

        rule = f'{driver_str}{stage_str}{injection_command_str}{approvers_str}{reason_code_str}'.encode('utf-8')
        rule = base64.b64encode(rule)

        vault_id = self._api.websdk.SecretStore.Add.post(
            base_64_data=rule.decode('utf-8'),
            vault_type=VaultTypes.blob,
            owner=workflow.dn,
            keyname=KeyNames.software_default,
            namespace=Namespaces.config
        ).vault_id

        self._api.websdk.Config.AddValue.post(
            object_dn=workflow.dn,
            attribute_name=WorkflowAttributes.rule_vault_id,
            value=vault_id
        )

        return workflow


@feature('Adapatable Workflow')
class AdaptableWorkflow(_WorkflowBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', stage: int, powershell_script_name: str,
               powershell_script_content: bytes, approvers: 'List[Union[Identity.Identity, str]]' = None, reason_code: int = None,
               use_approvers_from_powershell_script: bool = False, attributes: dict = None, get_if_already_exists: bool = True):
        """
        .. note::
            If a list of approver identity objects is provided, they will be added directly to the workflow as approvers of the 
            workflow. If the approvers should come from the PowerShell script, do not supply this parameter. Instead, set 
            ``use_approvers_from_powershell_script = True``. If the approvers should come from the object requiring the workflow, 
            such as the certificate object, then do not supply ``approvers`` or ``use_approvers_from_powershell_script``.

        Args:
            name: Name of the workflow object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            stage: One of the valid stages that represent the certificate lifecycle.
            powershell_script_name: Name (not path) of the actual PowerShell script on the TPP server.
            powershell_script_content: Contents of the PowerShell script as bytes.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` for each approver.
            reason_code: Integer reason code.
            use_approvers_from_powershell_script: If ``True`` and no ``approvers`` is supplied, then set the
                workflow to use the approvers defined by the script.
            attributes: Additional attributes to apply to the workflow object.
            get_if_already_exists: If the objects already exists, it is modified according to these parameters. Else
                and exception is raised..
        Returns:
            :ref:`config_object` of the workflow object.
        """
        if approvers:
            wf_approvers = ','.join([a.prefixed_universal for a in approvers])
        elif use_approvers_from_powershell_script:
            wf_approvers = "$CONFIG[$SELFDN$,'Adaptable Workflow Approvers']$"
        else:
            wf_approvers = ''

        workflow = self._create(
            name=name,
            parent_folder=parent_folder,
            is_adaptable=True,
            stage=stage,
            approvers=wf_approvers,
            reason_code=reason_code,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )

        self._api.websdk.Config.AddValue.post(
            object_dn=workflow.dn,
            attribute_name=AdaptableWorkflowAttributes.powershell_script,
            value=powershell_script_name
        ).assert_valid_response()

        vault_id = self._api.websdk.SecretStore.Add.post(
            base_64_data=self._calculate_hash(powershell_script_content),
            keyname=KeyNames.software_default,
            vault_type=VaultTypes.blob,
            namespace=Namespaces.config,
            owner=workflow.dn
        ).vault_id

        self._api.websdk.Config.WriteDn.post(
            object_dn=workflow.dn,
            attribute_name=AdaptableWorkflowAttributes.powershell_script_hash_vault_id,
            values=[vault_id]
        ).assert_valid_response()

        return workflow

    @staticmethod
    def _calculate_hash(script_content: bytes):
        return base64.b64encode(
            hashlib.sha256(
                script_content.decode().encode('utf-32-le')
            ).hexdigest().encode()
        ).decode()


@dataclass
class RC:
    code: int 
    name: str 
    description: str


@feature('Reason Code')
class ReasonCode(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)

        self._workflow_dn = r'\VED\Workflow Tickets'

    def create(self, code: int, name: str, description: str):
        """
        Args:
            code: An integer code.
            name: Name of the result code.
            description: Purpose of the result code.

        Returns:
            A ``ReasonCode`` object with these properties
            
            * **code** *(int)* - An integer code.
            * **name** *(str)* - Name of the result code.
            * **description** *(str)* - Purpose of the result code.
        """
        result = self._api.websdk.Config.AddValue.post(
            object_dn=self._workflow_dn,
            attribute_name=WorkflowTicketAttributes.approval_reason,
            value=f'{code},{name},{description}'
        )
        result.assert_valid_response()

        return RC(name=name, code=code, description=description)

    def delete(self, code: int, name: str = None):
        """
        Deletes a result code. ``name`` is not required, but when supplied, in the
        event that multiple codes exist with the same integer value, only the ones
        having ``name`` will be deleted.

        Args:
            code: An integer code.
            name: Name of the result code.
        """
        result_codes = self._api.websdk.Config.ReadDn.post(
            object_dn=self._workflow_dn,
            attribute_name=WorkflowTicketAttributes.approval_reason
        ).values

        if not result_codes:
            return

        search_string = f'{code},{name}' if name else f'{code}'
        for rc in result_codes:
            if search_string in rc:
                result = self._api.websdk.Config.RemoveDnValue.post(
                    object_dn=self._workflow_dn,
                    attribute_name=WorkflowTicketAttributes.approval_reason,
                    value=rc
                )
                result.assert_valid_response()


@feature('Standard Workflow')
class StandardWorkflow(_WorkflowBase):
    def __init__(self, api):
        super().__init__(api=api)

    def create(self, name: str, parent_folder: 'Union[Config.Object, str]', stage: int, injection_command: str = None,
               application_class_name: str = None, approvers: 'List[Union[Identity.Identity, str]]' = None, macro: str = None,
               reason_code: int = None, attributes: dict = None, get_if_already_exists: bool = True):
        """
        One of ``injection_command`` or ``approvers`` must be given.

        If a list of approver identity objects is provided, they will be added directly to the workflow as approvers of the
        workflow. If the approvers should come from the object requiring the workflow, such as the certificate object, then
        do not supply ``approvers``. If the approvers come from a TPP Macro, then supply ``macro`` with the desired macro.

        Args:
            name: Name of the workflow object.
            parent_folder: :ref:`config_object` or :ref:`dn` of the parent folder.
            stage: Workflow stage.
            injection_command: Command to be invoked on the target application.
            application_class_name: Application class name to trigger this workflow.
            approvers: List of :ref:`identity_object` or :ref:`prefixed_name` for each approver.
            macro: TPP approver macro.
            reason_code: Integer reason code.
            attributes: Additional attributes to apply to the workflow object.
            get_if_already_exists: If the objects already exists, it is modified according to these parameters. Else
                an exception is raised.

        Returns:
            :ref:`config_object` of the workflow object.
        """
        if approvers:
            wf_approvers = ','.join([a.prefixed_universal for a in approvers])
        elif macro:
            wf_approvers = macro
        else:
            wf_approvers = ''

        workflow = self._create(
            name=name,
            parent_folder=parent_folder,
            is_adaptable=False,
            stage=stage,
            injection_command=injection_command,
            application_class_name=str(application_class_name),
            approvers=wf_approvers,
            reason_code=reason_code,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )

        return workflow


@feature('Ticket')
class Ticket(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._workflow_ticket_dn = r'\VED\Workflow Tickets'

    @staticmethod
    def _validate_result_code(result):
        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.workflow_result)

    def create(self, obj: 'Union[Config.Object, str]', workflow: 'Union[Config.Object, str]',
               approvers: Union['List[Identity.Identity]', List[str]], reason: Union[RC, int, str],
               user_data: str = None):
        """
        Creates a workflow ticket on ``obj`` only if the object is in a state to received a workflow ticket.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object requiring a workflow ticket.
            workflow: :ref:`config_object` or :ref:`dn` of the workflow object.
            approvers:  List of :ref:`identity_object` or :ref:`prefixed_name` for each approver.
            reason: ``RC`` or integer reason code.
            user_data: User data.

        Returns:
            str: Workflow ticket name.
        """
        obj_dn = self._get_dn(obj)
        workflow_dn = self._get_dn(workflow)
        approver_prefixed_names = [
            {'PrefixedName': self._get_prefixed_name(a)} for a in approvers
        ]
        if isinstance(reason, RC):
            reason = reason.code
        response = self._api.websdk.Workflow.Ticket.Create.post(
            object_dn=obj_dn,
            workflow_dn=workflow_dn,
            approvers=approver_prefixed_names,
            reason=int(reason),
            user_data=user_data
        )
        self._validate_result_code(response.result)
        return response.guid

    def delete(self, ticket_name: str):
        """
        Deletes a workflow ticket.

        Args:
            ticket_name: Name of the workflow ticket.
        """
        result = self._api.websdk.Workflow.Ticket.Delete.post(guid=ticket_name).result
        self._validate_result_code(result)

    def details(self, ticket_name: str):
        """
        Returns the details of a ticket request.

        Args:
            ticket_name: Name of the workflow ticket.

        Returns:
            :class:`~.dataclasses.workflow.Details` of the workflow ticket.
        """
        response = self._api.websdk.Workflow.Ticket.Details.post(guid=ticket_name)
        self._validate_result_code(result=response.result)
        return response.details

    def exists(self, ticket_name: str):
        """
        Args:
            ticket_name: Name of the workflow ticket.

        Returns:
            bool: ``True`` when a particular workflow exists, otherwise ``False``.
        """
        result = self._api.websdk.Workflow.Ticket.Exists.post(guid=ticket_name).result
        return result.code == 1

    def get(self, obj: 'Union[Config.Object, str]' = None, user_data: str = None, expected_num_tickets: int = 1, timeout: int = 10):
        """
        Gets all tickets associated to ``obj``. If the minimum expected number of tickets do not
        appear on the ``obj``, then a warning is logged and whatever was found is returned and no
        error is raised.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object with a workflow ticket issued to it. If not given, only
                 ``user_data`` has effect.
            user_data: The string to filter results using the User Data attribute of the workflow ticket.
            expected_num_tickets: Minimum number of tickets expected to be written for the certificate.
            timeout: Time in seconds to wait for a ticket to be issued. Default is 10 seconds.

        Returns:
            List of :ref:`guid` for each ticket found.
        """
        if not obj:
            return self._api.websdk.Workflow.Ticket.Enumerate.post(
                user_data=user_data
            ).guids

        obj_dn = self._get_dn(obj)
        def get_tickets():
            return self._api.websdk.Workflow.Ticket.Enumerate.post(
                object_dn=obj_dn,
                user_data=user_data
            ).guids

        if timeout:
            tickets = []
            with self._Timeout(timeout=timeout) as to:
                while not to.is_expired():
                    tickets = get_tickets()
                    if len(tickets) >= expected_num_tickets:
                        return tickets

            self._log_warning_message(
                msg=f'The expected number of tickets on {obj_dn} was '
                    f'{expected_num_tickets}, but {len(tickets)} tickets were '
                    f'found instead.'
            )
            return tickets
        else:
            tickets = get_tickets()
            if len(tickets) < expected_num_tickets:
                self._log_warning_message(
                    msg=f'The expected number of tickets on {obj_dn} was '
                        f'{expected_num_tickets}, but {len(tickets)} tickets were '
                        f'found instead.'
                )
            return tickets

    def status(self, ticket_name: str):
        """
        Args:
            ticket_name: Name of the workflow ticket.

        Returns:
            str: The current status of a workflow ticket.
        """
        response = self._api.websdk.Workflow.Ticket.Status.post(guid=ticket_name)
        self._validate_result_code(result=response.result)
        return response.status

    def update_status(self, ticket_name: str, status: str, explanation: str = None, scheduled_start: datetime = None,
                      scheduled_stop: datetime = None):
        """
        Updates the status of a workflow ticket with the optional explanations and scheduled approvals. Marking a
        ticket as "Approved" will automatically delete the ticket.

        Args:
            ticket_name: Name of the workflow ticket.
            status: The new status of the workflow ticket.
            explanation: Reason for the new status.
            scheduled_start: Datetime to continue the approval process.
            scheduled_stop:  Datetime to expire the approval process.
        """
        result = self._api.websdk.Workflow.Ticket.UpdateStatus.post(
            guid=ticket_name,
            status=status,
            explanation=explanation,
            scheduled_start=scheduled_start.isoformat() if scheduled_start else None,
            scheduled_stop=scheduled_stop.isoformat() if scheduled_stop else None
        ).result
        self._validate_result_code(result)



from pytpp.properties.resultcodes import ResultCodes
from pytpp.tools.helpers.date_converter import from_date_string
from pytpp.properties.response_objects.dataclasses import flow


class Flow:
    @staticmethod
    def Result(code: int):
        return flow.Result(
            code=code,
            flow_result=ResultCodes.Flow.get(code, 'Unknown'),
        )

    @staticmethod
    def Ticket(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return flow.Ticket(
            approvals=[Flow.Approval(a) for a in response_object.get('Approvals')],
            approvers=response_object.get('Approvers'),
            creation_time=from_date_string(response_object.get('CreationTime')),
            environment=[Flow.KeyValue(i) for i in response_object.get('Environment')],
            flow_process_id=response_object.get('FlowProcessId'),
            id=response_object.get('Id'),
            identifier=response_object.get('Identifier'),
            product_code=response_object.get('ProductCode'),
            remaining_uses=response_object.get('RemainingUses'),
            required_approvals=response_object.get('RequiredApprovals'),
        )

    @staticmethod
    def Approval(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return flow.Approval(
            approval_time=from_date_string(response_object.get('ApprovalTime')),
            universal=response_object.get('Universal'),
        )

    @staticmethod
    def KeyValue(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return flow.KeyValue(
            key=response_object.get('Key'),
            value=response_object.get('Value'),
        )



from pytpp.properties.response_objects.dataclasses import identity


class Identity:
    @staticmethod
    def Identity(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return identity.Identity(
            full_name=response_object.get('FullName'),
            is_container=response_object.get('IsContainer'),
            is_group=response_object.get('IsGroup'),
            name=response_object.get('Name'),
            prefix=response_object.get('Prefix'),
            prefixed_name=response_object.get('PrefixedName'),
            prefixed_universal=response_object.get('PrefixedUniversal'),
            type=response_object.get('Type'),
            universal=response_object.get('Universal'),
        )

    @staticmethod
    def InvalidIdentity(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return identity.InvalidIdentity(
            prefix=response_object.get('Prefix'),
            prefixed_name=response_object.get('PrefixedName'),
            prefixed_universal=response_object.get('PrefixedUniversal'),
            universal=response_object.get('Universal'),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes


class GeotrustTrueFlexCAAttributes(CertificateAuthorityBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "GeotrustTrueFlex CA"
	certificate_block = Attribute('Certificate Block', min_version='15.3')
	certificate_transparency = Attribute('Certificate Transparency', min_version='16.3')
	server_type = Attribute('Server Type', min_version='15.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class PaloAltoNetworkFWAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Palo Alto Network FW"
	certificate_only = Attribute('Certificate Only', min_version='15.1')
	chain_cert = Attribute('Chain Cert', min_version='15.1')
	create_decryption_policy = Attribute('Create Decryption Policy', min_version='15.1')
	decryption_destinations = Attribute('Decryption Destinations', min_version='15.1')
	decryption_policy = Attribute('Decryption Policy', min_version='15.1')
	decryption_profile = Attribute('Decryption Profile', min_version='15.1')
	file_validation_disabled = Attribute('File Validation Disabled', min_version='15.1')
	lock_config = Attribute('Lock Config', min_version='15.1')
	network_validation_disabled = Attribute('Network Validation Disabled', min_version='15.1')
	replace_store = Attribute('Replace Store', min_version='15.1')
	tls_service_profile = Attribute('TLS Service Profile', min_version='20.4')
	temporary_name = Attribute('Temporary Name', min_version='15.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.agent_base import AgentBaseAttributes
from pytpp.attributes.top import TopAttributes


class AgentAttributes(AgentBaseAttributes, TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Agent"
	action = Attribute('Action')
	active_directory_dn_host_detail = Attribute('Active Directory DN Host Detail')
	active_directory_domain_host_detail = Attribute('Active Directory Domain Host Detail')
	active_directory_host_details = Attribute('Active Directory Host Details')
	active_directory_query_host_detail = Attribute('Active Directory Query Host Detail')
	active_directory_source_dn_host_detail = Attribute('Active Directory Source DN Host Detail')
	address = Attribute('Address')
	configuration_host_details = Attribute('Configuration Host Details')
	custom_host_detail = Attribute('Custom Host Detail')
	environment_host_details = Attribute('Environment Host Details')
	first_update = Attribute('First Update')
	host_os = Attribute('Host OS')
	hostname = Attribute('Hostname')
	last_update = Attribute('Last Update')
	os_id = Attribute('OS Id')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.top import TopAttributes


class SSHCertificateIssuanceFlowContainerAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "SSH Certificate Issuance Flow Container"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class CredentialBaseAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Credential Base"
	escalation_notice_interval = Attribute('Escalation Notice Interval')
	escalation_notice_start = Attribute('Escalation Notice Start')
	expiration = Attribute('Expiration')
	expiration_notice_interval = Attribute('Expiration Notice Interval')
	expiration_notice_start = Attribute('Expiration Notice Start')
	last_notification = Attribute('Last Notification')
	protection_key = Attribute('Protection Key')
	shared = Attribute('Shared')
	vault_id = Attribute('Vault Id')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.server_agent_base_device_placement_work import ServerAgentBaseDevicePlacementWorkAttributes


class ServerAgentSSHDevicePlacementWorkAttributes(ServerAgentBaseDevicePlacementWorkAttributes, metaclass=IterableMeta):
	__config_class__ = "Server Agent SSH Device Placement Work"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes
from pytpp.attributes.proxy import ProxyAttributes


class SymantecLHKCAAttributes(HTTPCABaseAttributes, ProxyAttributes, metaclass=IterableMeta):
	__config_class__ = "Symantec LHK CA"
	fields = Attribute('Fields')
	uri = Attribute('URI')



from typing import List 
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.metadata import Metadata
from pytpp.properties.response_objects.config import Config


class _Metadata(API):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Log')
        self.DefineItem = self._DefineItem(api_obj=api_obj)
        self.Find = self._Find(api_obj=api_obj)
        self.FindItem = self._FindItem(api_obj=api_obj)
        self.Get = self._Get(api_obj=api_obj)
        self.GetItemGuids = self._GetItemGuids(api_obj=api_obj)
        self.GetItems = self._GetItems(api_obj=api_obj)
        self.GetItemsForClass = self._GetItemsForClass(api_obj=api_obj)
        self.GetPolicyItems = self._GetPolicyItems(api_obj=api_obj)
        self.Items = self._Items(api_obj=api_obj)
        self.LoadItem = self._LoadItem(api_obj=api_obj)
        self.LoadItemGuid = self._LoadItemGuid(api_obj=api_obj)
        self.ReadEffectiveValues = self._ReadEffectiveValues(api_obj=api_obj)
        self.ReadPolicy = self._ReadPolicy(api_obj=api_obj)
        self.Set = self._Set(api_obj=api_obj)
        self.SetPolicy = self._SetPolicy(api_obj=api_obj)
        self.UndefineItem = self._UndefineItem(api_obj=api_obj)
        self.UpdateItem = self._UpdateItem(api_obj=api_obj)

    class _DefineItem(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/DefineItem')

        def post(self, item: dict):
            body = {
                'Item': item
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def dn(self) -> str:
                    return self._from_json('DN')

                @property
                @api_response_property()
                def item(self):
                    return Metadata.Item(self._from_json('Item'))

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _Find(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/Find')

        def post(self, item: str = None, item_guid: str = None, value: str = None):
            body = {
                'Item': item,
                'ItemGuid': item_guid,
                'Value': value
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def objects(self):
                    return [Config.Object(obj) for obj in self._from_json('Objects')]

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _FindItem(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/FindItem')

        def post(self, name: str):
            body = {
                'Name': name
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def item_guid(self) -> str:
                    return self._from_json('ItemGuid')

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _Get(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/Get')

        def post(self, dn: str, all_included: bool = None):
            body = {
                'DN': dn,
                'All': all_included
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def data(self):
                    return [Metadata.Data(self._from_json('Data'))]

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _GetItemGuids(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/GetItemGuids')

        def post(self, dn: str):
            body = {
                'DN': dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def item_guids(self) -> List[str]:
                    return self._from_json('ItemGuids')

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _GetItems(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/GetItems')

        def post(self, dn: str):
            body = {
                'DN': dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def items(self):
                    return [Metadata.Item(item) for item in self._from_json('Items')]

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _GetItemsForClass(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/GetItemsForClass')

        def post(self, config_class: str):
            body = {
                'ConfigClass': config_class
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def items(self):
                    return [Metadata.Item(item) for item in self._from_json('Items')]

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _GetPolicyItems(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/GetPolicyItems')

        def post(self, dn: str):
            body = {
                'DN': dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def policy_items(self):
                    return [Metadata.PolicyItem(item) for item in self._from_json('PolicyItems')]

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _Items(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/Items')

        def get(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def items(self):
                    return [Metadata.Item(item) for item in self._from_json('Items')]

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._get())

    class _LoadItem(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/LoadItem')

        def post(self, dn: str):
            body = {
                'DN': dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def item(self):
                    return Metadata.Item(self._from_json('Item'))

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _LoadItemGuid(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/LoadItemGuid')

        def post(self, dn: str):
            body = {
                'DN': dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def item_guid(self) -> str:
                    return self._from_json('ItemGuid')

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _ReadEffectiveValues(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/ReadEffectiveValues')

        def post(self, dn: str, item_guid: str):
            body = {
                'DN': dn,
                'ItemGuid': item_guid
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def policy_dn(self) -> str:
                    return self._from_json('PolicyDn')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

                @property
                @api_response_property()
                def values(self) -> List[str]:
                    return self._from_json('Values')

            return _Response(response=self._post(data=body))

    class _ReadPolicy(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/ReadPolicy')

        def post(self, dn: str, item_guid: str, obj_type: str):
            body = {
                'DN': dn,
                'ItemGuid': item_guid,
                'Type': obj_type
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

                @property
                @api_response_property()
                def values(self) -> List[str]:
                    return self._from_json('Values')

            return _Response(response=self._post(data=body))

    class _Set(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/Set')

        def post(self, dn: str, guid_data: list, keep_existing: bool = False):
            body = {
                'DN': dn,
                'GuidData': guid_data,
                'KeepExisting': keep_existing
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _SetPolicy(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/SetPolicy')

        def post(self, dn: str, config_class: str, guid_data: list, locked: bool = False):
            body = {
                'DN': dn,
                'ConfigClass': config_class,
                'GuidData': guid_data,
                'Locked': locked
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _UndefineItem(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/UndefineItem')

        def post(self, item_guid: str, remove_data: bool = True):
            body = {
                'ItemGuid': item_guid,
                'RemoveData': remove_data
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))

    class _UpdateItem(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url=f'/Metadata/UpdateItem')

        def post(self, item: dict = None, update: dict = None):
            body = {
                'ItemGuid': item,
                'Update': update
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def locked(self) -> bool:
                    return self._from_json('Locked')

                @property
                @api_response_property()
                def result(self):
                    return Metadata.Result(self._from_json('Result'))

            return _Response(response=self._post(data=body))



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class VAMnShieldAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "VAM nShield"
	file_validation_disabled = Attribute('File Validation Disabled', min_version='15.3')
	install_path = Attribute('Install Path')
	km_local_path = Attribute('KM Local Path')
	key_store_validation_disabled = Attribute('Key Store Validation Disabled')
	module_id = Attribute('Module Id')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class LayoutRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Layout Root"
	teams_template_code_signing_vault_id = Attribute('Teams Template Code Signing Vault Id', min_version='20.1')
	teams_template_ssh_vault_id = Attribute('Teams Template SSH Vault Id', min_version='20.1')
	teams_template_tls_vault_id = Attribute('Teams Template TLS Vault Id', min_version='20.1')



from pytpp.tools.deprecation import DeprecationMeta
from pytpp.properties.config import (
    ApplicationGroupClassNames, ApplicationClassNames, CertificateClassNames, DevicesClassNames, DiscoveryClassNames,
    IdentityClassNames, PlacementRulesClassNames, PlatformsClassNames, WorkflowClassNames, ClientWorkClassNames,
    ClientGroupsClassNames, FolderClassNames, CertificateAuthorityClassNames
)

class Classes(metaclass=DeprecationMeta):
    __deprecation_reason__ = 'Using Classes will be deprecated soon. Get the TPP class names from the attributes using ' \
                             'pytpp.Attributes (from pytpp import Attributes) instead. Alternatively you can import ' \
                             'pytpp.ClassNames, which provides all class names in TPP whereas the Attributes only ' \
                             'provide the commonly used TPP class names.'
    Application = ApplicationClassNames
    ApplicationGroup = ApplicationGroupClassNames
    Certificate = CertificateClassNames
    CertificateAuthority = CertificateAuthorityClassNames
    ClientGroups = ClientGroupsClassNames
    ClientWork = ClientWorkClassNames
    Device = DevicesClassNames
    Discovery = DiscoveryClassNames
    Folder = FolderClassNames
    Identity = IdentityClassNames
    PlacementRules = PlacementRulesClassNames
    Platforms = PlatformsClassNames
    Workflow = WorkflowClassNames



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class ServiceModuleAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Service Module"
	driver_arguments = Attribute('Driver Arguments')
	driver_name = Attribute('Driver Name')
	heartbeat_interval = Attribute('Heartbeat Interval')
	interval = Attribute('Interval')
	started_by = Attribute('Started By')



from pytpp.attributes._helper import IterableMeta, Attribute


class AgentBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Agent Base"
	agent_guid = Attribute('Agent GUID')
	location = Attribute('Location')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class FlowActionBaseAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Flow Action Base"
	rank = Attribute('Rank', min_version='19.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class ExclusionAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Exclusion"
	rule = Attribute('Rule')



from pytpp.api.api_base import API, APIResponse


class _Revoke:
    def __init__(self, api_obj):
        self.Token = self._Token(api_obj=api_obj)

    class _Token(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Revoke/Token')
            self._url = self._url.replace('vedsdk', 'vedauth')

        def get(self):
            return APIResponse(response=self._get())



import time
import os
import re
from typing import TYPE_CHECKING
from pytpp.features.definitions.exceptions import InvalidResultCode, ObjectDoesNotExist
from pytpp.properties.response_objects.config import Config
from pytpp.properties.response_objects.identity import Identity
from pytpp.properties.response_objects.dataclasses.config import Object as ConfigObject
from pytpp.properties.response_objects.dataclasses.identity import Identity
from pytpp.tools.logger import logger, LogTags
from pytpp.properties.secret_store import Namespaces
from typing import List, Dict, Union
from packaging.version import Version
if TYPE_CHECKING:
    from pytpp.api.authenticate import Authenticate


def feature(name: str):
    def decorate(cls):
        if int(os.getenv('PYTPP_DOC_IN_PROGRESS', 0)):
            return cls
        setattr(cls, '__feature__', name)
        return logger.wrap_class(
            log_tag=LogTags.feature,
            func_regex_exclude='_.*'
        )(cls)
    return decorate


class FeatureBase:
    def __init__(self, api: 'Authenticate'):
        self._api = api

    def _config_create(self, name: str, parent_folder_dn: str, config_class: str, attributes: dict = None,
                       get_if_already_exists: bool = True, keep_list_values: bool = False):
        if attributes:
            attributes = self._name_value_list(attributes=attributes, keep_list_values=keep_list_values)

        dn = f'{parent_folder_dn}\\{name}'
        response = self._api.websdk.Config.Create.post(object_dn=dn, class_name=str(config_class),
                                                 name_attribute_list=attributes or [])
        result = response.result
        if result.code != 1:
            if result.code == 401 and get_if_already_exists:
                return self._get_config_object(object_dn=dn)
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

        return response.object

    def _config_delete(self, object_dn, recursive: bool = False):
        result = self._api.websdk.Config.Delete.post(object_dn=object_dn, recursive=recursive).result
        if result.code != 1:
            raise InvalidResultCode(code=result.code, code_description=result.config_result)

    def _get_config_object(self, object_dn: str = None, object_guid: str = None,
                           raise_error_if_not_exists: bool = True, valid_class_names: List[str] = None):
        if not (object_dn or object_guid):
            raise ValueError(
                'Must supply either an Object DN or Object GUID, but neither was provided.'
            )
        obj = Config.Object({})
        if isinstance(object_dn, ConfigObject):
            obj = object_dn
        elif isinstance(object_guid, ConfigObject):
            obj = object_guid
        else:
            response = self._api.websdk.Config.IsValid.post(object_dn=object_dn, object_guid=object_guid)
            if response.result.code == 400:
                if raise_error_if_not_exists:
                    raise ObjectDoesNotExist(f'"{object_dn or object_guid}" does not exist.')
            else:
                obj = response.object
        if valid_class_names and obj.type_name not in valid_class_names:
            valid_class_names = "\n".join(f"*  {i}" for i in valid_class_names)
            raise TypeError(f'"{object_dn or object_guid}" exists, but is not the expected class type.\n'
                             f'Got type "{obj.type_name}" instead of one of \n{valid_class_names}.')
        return obj

    def _get_identity_object(self, prefixed_name: str = None, prefixed_universal: str = None,
                             raise_error_if_not_exists: bool = True):
        if not (prefixed_name or prefixed_universal):
            raise ValueError(
                'Must supply either an prefixed_name or prefixed_universal, but neither was provided.'
            )
        if isinstance(prefixed_name, Identity):
            return prefixed_name
        if isinstance(prefixed_universal, Identity):
            return prefixed_universal

        result = self._api.websdk.Identity.Validate.post(
            identity=self._identity_dict(prefixed_name=prefixed_name, prefixed_universal=prefixed_universal)
        )
        if result.is_valid_response() and result.api_response.content:
            identity = result.identity
        elif raise_error_if_not_exists:
            target = prefixed_name or prefixed_universal
            raise ObjectDoesNotExist(f'Could not find identity "{target}".')
        else:
            identity = Identity.Identity(response_object={})
        return identity

    @staticmethod
    def _identity_dict(prefixed_name: str = None, prefixed_universal: str = None):
        """
        Creates an ID object to write to the Identity APIs.

        Args:
            prefixed_name: The prefixed name of the Identity object.
            prefixed_universal: The prefixed universal GUID of the Identity object.

        Returns:
            {
                'PrefixedUniversal': ``prefixed_universal``,
                'PrefixedName': ``prefixed_name``
            }
        """
        d = {}
        if prefixed_name:
            d.update({'PrefixedName': prefixed_name})
        if prefixed_universal:
            d.update({'PrefixedUniversal': prefixed_universal})
        return d

    @staticmethod
    def _is_prefixed_universal(identity):
        if isinstance(identity, str):
            if ':' not in identity:
                return False
            prefix, identity = identity.split(':', maxsplit=1)
            regex = '^[{]?[0-9a-fA-F]{8}-([0-9a-fA-F]{4}-){3}[0-9a-fA-F]{12}[}]?$'
            return bool(re.match(pattern=regex, string=identity))
        return False

    @staticmethod
    def _is_obj_guid(obj: str):
        regex = '^[{]?[0-9a-fA-F]{8}-([0-9a-fA-F]{4}-){3}[0-9a-fA-F]{12}[}]?$'
        return isinstance(obj, str) and bool(re.match(pattern=regex, string=obj))

    def _get_prefixed_name(self, identity: 'Union[Identity, str]'):
        if hasattr(identity, 'prefixed_name'):
            return identity.prefixed_name
        if self._is_prefixed_universal(identity):
            return self._get_identity_object(prefixed_universal=identity).prefixed_name
        return identity

    def _get_prefixed_universal(self, identity: 'Union[Identity, str]'):
        if hasattr(identity, 'prefixed_universal'):
            return identity.prefixed_universal
        if self._is_prefixed_universal(identity):
            return identity
        return self._get_identity_object(prefixed_name=identity).prefixed_universal

    def _get_dn(self, obj: 'Union[Config.Object, str]', parent_dn: str = None):
        if hasattr(obj, 'dn'):
            return obj.dn
        if self._is_obj_guid(obj):
            return self._get_config_object(object_guid=obj).dn
        if parent_dn and not obj.startswith(parent_dn):
            obj = parent_dn.rstrip('\\') + '\\' + obj.rstrip('\\')
        if not obj.startswith(r'\VED'):
            return '\\VED\\' + obj.strip("\\")
        else:
            return obj

    def _get_guid(self, obj: 'Union[Config.Object, str]', parent_dn: str = None):
        if hasattr(obj, 'guid'):
            return obj.guid
        if self._is_obj_guid(obj):
            return obj
        if parent_dn and not obj.startswith(parent_dn):
            obj = parent_dn.rstrip('\\') + f'\\{obj}'
        if not obj.startswith(r'\VED'):
            obj = '\\VED\\' + obj.strip("\\")
        return self._get_config_object(object_dn=obj).guid

    @staticmethod
    def _log_warning_message(msg: str):
        logger.warning(msg=msg, num_prev_callers=2)

    @staticmethod
    def __no_op(*args, **kwargs):
        pass

    @staticmethod
    def _name_type_value(name: str, type: str, value):
        return {'Name': str(name), 'Type': str(type), 'Value': str(value)}

    @staticmethod
    def _name_value_list(attributes: Dict[str, List[str]], keep_list_values: bool = False):
        nvl = []
        for name, value in attributes.items():
            if value is None:
                continue
            elif isinstance(value, list):
                if keep_list_values is True:
                    nvl.append({'Name': str(name), 'Value': value})
                else:
                    for v in value:
                        nvl.append({'Name': str(name), 'Value': str(v)})
            elif not isinstance(value, dict):
                nvl.append({'Name': str(name), 'Value': str(value)})
        return nvl

    def _secret_store_delete(self, object_dn: str, namespace: str = Namespaces.config):
        result = self._api.websdk.SecretStore.OwnerDelete.post(namespace=namespace, owner=object_dn).result
        if result.code != 0:
            raise InvalidResultCode(code=result.code, code_description=result.secret_store_result)

    def _is_version_compatible(self, minimum: str = '', maximum: str = ''):
        if minimum and self._api._tpp_version <= Version(minimum):
            logger.error(f'Incompatible version. This feature requires at least TPP {minimum}.')
            return False
        if maximum and self._api._tpp_version >= Version(maximum):
            logger.error(f'Incompatible version. This feature is no longer available after TPP {maximum}.')
            return False
        return True

    class _Timeout:
        def __init__(self, timeout):
            self.timeout = timeout
            self.max_time = timeout + time.time()

        def __enter__(self):
            logger.set_rule(
                log_tag=LogTags.feature,
                blacklist_function=lambda x: x is LogTags.api,
                why=f'Disabling all logs during timeout to reduce redundant logging.'
            )
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            logger.set_rule(
                log_tag=LogTags.feature,
                reset=True,
                why=f'Enabling all logs after timeout.'
            )
            return

        def is_expired(self, poll: float = 0.5):
            time.sleep(poll)
            return time.time() >= self.max_time



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.credential_base import CredentialBaseAttributes


class GenerationalCredentialAttributes(CredentialBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Generational Credential"
	last = Attribute('Last')
	visibility = Attribute('Visibility')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class PKCS12Attributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "PKCS#12"
	bundle_certificate = Attribute('Bundle Certificate')
	certificate_chain_file = Attribute('Certificate Chain File')
	certificate_file = Attribute('Certificate File')
	create_store = Attribute('Create Store', min_version='15.1')
	file_validation_disabled = Attribute('File Validation Disabled')
	friendly_name = Attribute('Friendly Name')
	network_validation_disabled = Attribute('Network Validation Disabled')
	recycle_alias = Attribute('Recycle Alias', min_version='15.1')
	replace_store = Attribute('Replace Store', min_version='15.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.credential_base import CredentialBaseAttributes


class UsernamePasswordCredentialAttributes(CredentialBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Username Password Credential"
	username = Attribute('Username')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class LogRootAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Log Root"
	log_application_container = Attribute('Log Application Container')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class MetadataCategoryAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Metadata Category"
	label_text = Attribute('Label Text')
	localization = Attribute('Localization')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.bulk_application_base import BulkApplicationBaseAttributes


class AdaptableBulkAppAttributes(BulkApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Adaptable Bulk App"
	log_debug = Attribute('Log Debug', min_version='19.3')
	option_1 = Attribute('Option 1', min_version='18.3')
	option_2 = Attribute('Option 2', min_version='18.3')
	pk_credential = Attribute('PK Credential', min_version='18.3')
	password_1 = Attribute('Password 1', min_version='18.3')
	powershell_script = Attribute('PowerShell Script', min_version='18.3')
	powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id', min_version='18.4')
	script_execution_timeout = Attribute('Script Execution Timeout', min_version='20.2')
	script_hash_mismatch_error = Attribute('Script Hash Mismatch Error', min_version='18.4')
	secondary_credential = Attribute('Secondary Credential', min_version='18.3')
	text_field_1 = Attribute('Text Field 1', min_version='18.3')
	text_field_2 = Attribute('Text Field 2', min_version='18.3')
	text_field_3 = Attribute('Text Field 3', min_version='18.3')
	text_field_4 = Attribute('Text Field 4', min_version='18.3')
	text_field_5 = Attribute('Text Field 5', min_version='18.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes
from pytpp.attributes.proxy import ProxyAttributes


class RedhatCAAttributes(HTTPCABaseAttributes, ProxyAttributes, metaclass=IterableMeta):
	__config_class__ = "Redhat CA"
	agent_port = Attribute('Agent Port')
	agent_url_surffix = Attribute('Agent URL Surffix')
	end_entity_port = Attribute('End Entity Port')
	end_entity_url_surffix = Attribute('End Entity URL Surffix')
	use_profile = Attribute('Use Profile')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_certificate_work import ClientCertificateWorkAttributes


class NetworkDeviceCertificateWorkAttributes(ClientCertificateWorkAttributes, metaclass=IterableMeta):
	__config_class__ = "Network Device Certificate Work"
	authentication_credentials = Attribute('Authentication Credentials', min_version='20.1')
	ca_template_trust_anchors_enabled = Attribute('CA Template Trust Anchors Enabled', min_version='20.1')
	certificates_distribution_type = Attribute('Certificates Distribution Type', min_version='20.1')
	client_certificate_eku_checks_enabled = Attribute('Client Certificate Eku Checks Enabled', min_version='20.1')
	explicit_trust_anchors = Attribute('Explicit Trust Anchors', min_version='20.1')
	fallback_to_http_auth = Attribute('Fallback To Http Auth', min_version='20.1')
	http_basic_auth_disabled = Attribute('Http Basic Auth Disabled', min_version='20.1')
	http_digest_auth_disabled = Attribute('Http Digest Auth Disabled', min_version='20.1')
	pop_mode = Attribute('PoP Mode', min_version='20.1')
	reenrollment_subset_subject_matching_disabled = Attribute('ReEnrollment Subset Subject Matching Disabled', min_version='20.1')
	require_additional_http_auth = Attribute('Require Additional Http Auth', min_version='20.1')
	retryafter_disabled = Attribute('RetryAfter Disabled', min_version='20.1')
	revocation_check_timeout = Attribute('Revocation Check Timeout', min_version='20.1')
	revocation_mode = Attribute('Revocation Mode', min_version='20.1')
	revoke_existing_certificate_on_reenrollment = Attribute('Revoke Existing Certificate On ReEnrollment', min_version='20.1')
	revoke_existing_certificate_on_reenrollment_delay = Attribute('Revoke Existing Certificate On ReEnrollment Delay', min_version='20.1')
	use_implicit_trust_anchors = Attribute('Use Implicit Trust Anchors', min_version='20.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class AmazonCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Amazon CA"
	access_key_id = Attribute('Access Key ID', min_version='16.1')
	amazon_private_ca = Attribute('Amazon Private CA', min_version='21.2')
	aws_credential_dn = Attribute('Aws Credential DN', min_version='18.3')
	domain_control_validation = Attribute('Domain Control Validation', min_version='16.1')
	interval = Attribute('Interval', min_version='16.1')
	keys_generated_by_the_ca = Attribute('Keys Generated By The CA', min_version='21.2')
	region_code = Attribute('Region Code', min_version='16.1')
	retrieval_period = Attribute('Retrieval Period', min_version='16.1')
	secret_access_key = Attribute('Secret Access Key', min_version='16.1')
	signature_algorithm = Attribute('Signature Algorithm', min_version='21.2')
	use_private_ca = Attribute('Use Private CA', min_version='21.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_work_base import ClientWorkBaseAttributes


class ClientCertificateDiscoveryWorkAttributes(ClientWorkBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Certificate Discovery Work"
	certificate_location_dn = Attribute('Certificate Location DN', min_version='15.2')
	certificate_scanner_capi = Attribute('Certificate Scanner CAPI')
	certificate_scanner_map = Attribute('Certificate Scanner Map')
	certificate_scanner_native = Attribute('Certificate Scanner Native')
	certificate_scanner_native_stores = Attribute('Certificate Scanner Native Stores')
	certificate_scanner_path = Attribute('Certificate Scanner Path')
	clear_cache_timestamp = Attribute('Clear Cache Timestamp', min_version='20.1')
	credential = Attribute('Credential')
	days_of_month = Attribute('Days Of Month')
	days_of_week = Attribute('Days Of Week')
	device_location_dn = Attribute('Device Location DN', min_version='15.2')
	discovery_container = Attribute('Discovery Container')
	exclude_remote_mount_points = Attribute('Exclude Remote Mount Points')
	interval = Attribute('Interval')
	log_threshold = Attribute('Log Threshold')
	max_filesize = Attribute('Max Filesize')
	minimize_resource_use = Attribute('Minimize Resource Use')
	naming_pattern = Attribute('Naming Pattern')
	placement_rule = Attribute('Placement Rule', min_version='15.2')
	placement_summary = Attribute('Placement Summary', min_version='15.2')
	schedule_type = Attribute('Schedule Type')
	start_time = Attribute('Start Time')
	treat_unknown_roots_as_self_signed = Attribute('Treat Unknown Roots As Self Signed', min_version='21.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.driver_base import DriverBaseAttributes
from pytpp.attributes.monitoring_base import MonitoringBaseAttributes


class X509CertificateBaseAttributes(DriverBaseAttributes, MonitoringBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "X509 Certificate Base"
	adaptable_ca_binary_data_vault_id = Attribute('Adaptable CA:Binary Data Vault ID', min_version='16.2')
	adaptable_ca_early_password_vault_id = Attribute('Adaptable CA:Early Password Vault ID', min_version='17.4')
	adaptable_ca_early_pkcs7_vault_id = Attribute('Adaptable CA:Early Pkcs7 Vault ID', min_version='16.2')
	adaptable_ca_early_private_key_vault_id = Attribute('Adaptable CA:Early Private Key Vault ID', min_version='17.4')
	adaptable_ca_script_hash_mismatch_error = Attribute('Adaptable CA:Script Hash Mismatch Error', min_version='18.4')
	adaptable_workflow_approvers = Attribute('Adaptable Workflow Approvers', min_version='18.3')
	adaptable_workflow_reference_id = Attribute('Adaptable Workflow Reference ID', min_version='18.3')
	adaptable_workflow_stage = Attribute('Adaptable Workflow Stage', min_version='18.3')
	address = Attribute('Address')
	allow_private_key_reuse = Attribute('Allow Private Key Reuse', min_version='15.4')
	amazon_ca_first_pickup_request = Attribute('Amazon CA:First Pickup Request', min_version='16.1')
	amazon_ca_timestamp = Attribute('Amazon CA:Timestamp', min_version='16.1')
	approved_issuer = Attribute('Approved Issuer')
	approver = Attribute('Approver')
	csr_thumbprint = Attribute('CSR Thumbprint', min_version='20.1')
	csr_vault_id = Attribute('CSR Vault Id')
	certificate_authority = Attribute('Certificate Authority')
	certificate_download_pbes2_algorithm = Attribute('Certificate Download: PBES2 Algorithm', min_version='16.3')
	certificate_process_validator = Attribute('Certificate Process Validator', min_version='19.3')
	certificate_vault_id = Attribute('Certificate Vault Id')
	city = Attribute('City')
	comodo_ca_dcv_email = Attribute('Comodo CA:DCV Email')
	comodo_ca_server_type_id = Attribute('Comodo CA:Server Type Id')
	comodo_ccm_ca_pass_phrase = Attribute('Comodo CCM CA:Pass Phrase')
	comodo_ccm_ca_server_type = Attribute('Comodo CCM CA:Server Type')
	consumers = Attribute('Consumers')
	country = Attribute('Country')
	created_by = Attribute('Created By', min_version='15.1')
	digicert_ca_address = Attribute('DigiCert CA:Address')
	digicert_ca_request_id = Attribute('DigiCert CA:Request Id')
	digicert_ca_server_type = Attribute('DigiCert CA:Server Type')
	digicert_ca_specific_end_date = Attribute('DigiCert CA:Specific End Date', min_version='17.3')
	digicert_ca_zip = Attribute('DigiCert CA:Zip')
	disable_automatic_renewal = Attribute('Disable Automatic Renewal')
	disable_password_complexity = Attribute('Disable Password Complexity', min_version='15.4')
	discovered_by_dn = Attribute('Discovered By DN', min_version='19.1')
	discovered_on = Attribute('Discovered On')
	domain_suffix_whitelist = Attribute('Domain Suffix Whitelist', min_version='15.4')
	esm_ca_override_default_key_update_policy = Attribute('ESM CA:Override Default Key Update Policy')
	est_reenrollment_in_progress = Attribute('EST ReEnrollment In Progress', min_version='20.1')
	elliptic_curve = Attribute('Elliptic Curve', min_version='17.1')
	encryption_driver = Attribute('Encryption Driver')
	enforce_unique_subject = Attribute('Enforce Unique Subject', min_version='15.4')
	entrust_pki_gateway_early_private_key_vault_id = Attribute('Entrust PKI Gateway:Early Private Key Vault ID', min_version='20.2')
	entrust_pki_gateway_early_x509_vault_id = Attribute('Entrust PKI Gateway:Early X509 Vault ID', min_version='20.2')
	entrustnet_ca_additional_emails = Attribute('EntrustNET CA:Additional Emails', min_version='17.1')
	entrustnet_ca_additional_field_value = Attribute('EntrustNET CA:Additional Field Value')
	entrustnet_ca_email_address = Attribute('EntrustNET CA:Email Address')
	entrustnet_ca_first_name = Attribute('EntrustNET CA:First Name')
	entrustnet_ca_first_pickup_request = Attribute('EntrustNET CA:First Pickup Request', min_version='15.4')
	entrustnet_ca_last_name = Attribute('EntrustNET CA:Last Name')
	entrustnet_ca_specific_end_date = Attribute('EntrustNET CA:Specific End Date')
	entrustnet_ca_timestamp = Attribute('EntrustNET CA:Timestamp', min_version='15.4')
	escalation_notice_interval = Attribute('Escalation Notice Interval')
	escalation_notice_start = Attribute('Escalation Notice Start')
	expiration_notice_interval = Attribute('Expiration Notice Interval')
	expiration_notice_start = Attribute('Expiration Notice Start')
	fields = Attribute('Fields')
	generate_keypair_on_application = Attribute('Generate Keypair On Application')
	geotrust_ca_address = Attribute('GeoTrust CA:Address')
	geotrust_ca_admin_contact_address = Attribute('GeoTrust CA:Admin Contact Address')
	geotrust_ca_admin_contact_city = Attribute('GeoTrust CA:Admin Contact City')
	geotrust_ca_admin_contact_country = Attribute('GeoTrust CA:Admin Contact Country')
	geotrust_ca_admin_contact_email_address = Attribute('GeoTrust CA:Admin Contact Email Address')
	geotrust_ca_admin_contact_first_name = Attribute('GeoTrust CA:Admin Contact First Name')
	geotrust_ca_admin_contact_last_name = Attribute('GeoTrust CA:Admin Contact Last Name')
	geotrust_ca_admin_contact_organization = Attribute('GeoTrust CA:Admin Contact Organization')
	geotrust_ca_admin_contact_phone_number = Attribute('GeoTrust CA:Admin Contact Phone Number')
	geotrust_ca_admin_contact_postal_code = Attribute('GeoTrust CA:Admin Contact Postal Code')
	geotrust_ca_admin_contact_state = Attribute('GeoTrust CA:Admin Contact State')
	geotrust_ca_admin_contact_title = Attribute('GeoTrust CA:Admin Contact Title')
	geotrust_ca_authentication_comments = Attribute('GeoTrust CA:Authentication Comments')
	geotrust_ca_authentication_statuses = Attribute('GeoTrust CA:Authentication Statuses')
	geotrust_ca_enrollment_mode = Attribute('GeoTrust CA:Enrollment Mode', min_version='15.1')
	geotrust_ca_order_id = Attribute('GeoTrust CA:Order Id')
	geotrust_ca_postal_code = Attribute('GeoTrust CA:Postal Code')
	geotrust_ca_server_count = Attribute('GeoTrust CA:Server Count')
	geotrust_ca_server_type = Attribute('GeoTrust CA:Server Type')
	geotrust_ca_telephone_number = Attribute('GeoTrust CA:Telephone Number')
	geotrust_ca_timestamp = Attribute('GeoTrust CA:Timestamp')
	geotrust_enterprise_ca_address = Attribute('GeoTrust Enterprise CA:Address')
	geotrust_enterprise_ca_admin_contact_email_address = Attribute('GeoTrust Enterprise CA:Admin Contact Email Address')
	geotrust_enterprise_ca_admin_contact_first_name = Attribute('GeoTrust Enterprise CA:Admin Contact First Name')
	geotrust_enterprise_ca_admin_contact_last_name = Attribute('GeoTrust Enterprise CA:Admin Contact Last Name')
	geotrust_enterprise_ca_admin_contact_phone_number = Attribute('GeoTrust Enterprise CA:Admin Contact Phone Number')
	geotrust_enterprise_ca_admin_contact_title = Attribute('GeoTrust Enterprise CA:Admin Contact Title')
	geotrust_enterprise_ca_approver_email = Attribute('GeoTrust Enterprise CA:Approver Email')
	geotrust_enterprise_ca_enrollment_mode = Attribute('GeoTrust Enterprise CA:Enrollment Mode', min_version='15.1')
	geotrust_enterprise_ca_postal_code = Attribute('GeoTrust Enterprise CA:Postal Code')
	geotrust_enterprise_ca_server_count = Attribute('GeoTrust Enterprise CA:Server Count')
	geotrust_enterprise_ca_server_type = Attribute('GeoTrust Enterprise CA:Server Type')
	geotrust_enterprise_ca_telephone_number = Attribute('GeoTrust Enterprise CA:Telephone Number')
	geotrust_enterprise_ca_timestamp = Attribute('GeoTrust Enterprise CA:Timestamp')
	given_name = Attribute('Given Name')
	globalsign_mssl_ca_email = Attribute('GlobalSign MSSL CA:Email')
	globalsign_mssl_ca_first_name = Attribute('GlobalSign MSSL CA:First Name')
	globalsign_mssl_ca_last_name = Attribute('GlobalSign MSSL CA:Last Name')
	globalsign_mssl_ca_phone = Attribute('GlobalSign MSSL CA:Phone')
	google_cloud_ca_early_x509_chain_vault_id = Attribute('Google Cloud CA:Early X509 Chain Vault ID', min_version='21.3')
	google_cloud_ca_early_x509_vault_id = Attribute('Google Cloud CA:Early X509 Vault ID', min_version='21.3')
	grouping_id = Attribute('Grouping Id')
	in_error = Attribute('In Error')
	in_process = Attribute('In Process', min_version='15.3')
	internet_email_address = Attribute('Internet EMail Address')
	issued_to = Attribute('Issued To')
	key_algorithm = Attribute('Key Algorithm', min_version='17.1')
	key_bit_strength = Attribute('Key Bit Strength')
	key_storage_location = Attribute('Key Storage Location', min_version='19.2')
	keynectis_sequoia_ca_fields = Attribute('Keynectis Sequoia CA:Fields')
	last_evaluated_on = Attribute('Last Evaluated On', min_version='17.2')
	last_notification = Attribute('Last Notification')
	last_renewed_by = Attribute('Last Renewed By', min_version='16.2')
	last_renewed_on = Attribute('Last Renewed On', min_version='16.2')
	last_validation_state_update = Attribute('Last Validation State Update', min_version='16.3')
	license_count = Attribute('License Count')
	management_type = Attribute('Management Type')
	manual_approval = Attribute('Manual Approval')
	manual_csr = Attribute('Manual Csr')
	microsoft_ca_request_approved = Attribute('Microsoft CA:Request Approved')
	microsoft_ca_specific_end_date = Attribute('Microsoft CA:Specific End Date', min_version='18.2')
	network_validation_disabled = Attribute('Network Validation Disabled')
	opentrust_pki_ca_fields = Attribute('OpenTrust PKI CA:Fields')
	opentrust_pki_ca_first_pickup_request = Attribute('OpenTrust PKI CA:First Pickup Request')
	opentrust_pki_ca_requester_email = Attribute('OpenTrust PKI CA:Requester Email')
	options = Attribute('Options')
	organization = Attribute('Organization')
	organizational_unit = Attribute('Organizational Unit')
	origin = Attribute('Origin', min_version='19.1')
	pkcs10_hash_algorithm = Attribute('PKCS10 Hash Algorithm', min_version='15.1')
	postal_code = Attribute('Postal Code')
	private_key_vault_id = Attribute('Private Key Vault Id')
	prohibit_wildcard = Attribute('Prohibit Wildcard', min_version='15.4')
	prohibited_subject_attributes = Attribute('Prohibited Subject Attributes', min_version='19.1')
	protection_key = Attribute('Protection Key')
	public_key_vault_id = Attribute('Public Key Vault Id', min_version='19.4')
	renewal_window = Attribute('Renewal Window')
	reverse_dc_order = Attribute('Reverse DC Order', min_version='19.1')
	revocation_check_disabled = Attribute('Revocation Check Disabled')
	revocation_check_in_error = Attribute('Revocation Check In Error')
	revocation_check_last_checked = Attribute('Revocation Check Last Checked', min_version='16.2')
	revocation_check_now = Attribute('Revocation Check Now')
	revocation_check_status = Attribute('Revocation Check Status')
	revocation_original_request = Attribute('Revocation Original Request', min_version='17.2')
	revocation_request = Attribute('Revocation Request')
	scep_transaction_id = Attribute('Scep Transaction Id')
	server_type = Attribute('Server Type')
	signing_request_subject = Attribute('Signing Request Subject')
	specific_end_date = Attribute('Specific End Date', min_version='20.2')
	stage = Attribute('Stage')
	state = Attribute('State')
	status = Attribute('Status')
	surname = Attribute('Surname')
	symantec_lhk_ca_fields = Attribute('Symantec LHK CA:Fields')
	telephone = Attribute('Telephone')
	thawte_ca_emails = Attribute('Thawte CA:Emails', min_version='15.4')
	thawte_ca_enrollment_mode = Attribute('Thawte CA:Enrollment Mode', min_version='15.4')
	thawte_ca_first_pickup_request = Attribute('Thawte CA:First Pickup Request', min_version='15.4')
	thawte_ca_timestamp = Attribute('Thawte CA:Timestamp', min_version='15.4')
	transaction_id = Attribute('Transaction Id')
	trusted_status = Attribute('Trusted Status')
	trustwave_ca_enrollment_mode = Attribute('Trustwave CA:Enrollment Mode')
	trustwave_ca_first_pickup_request = Attribute('Trustwave CA:First Pickup Request')
	trustwave_ca_timestamp = Attribute('Trustwave CA:Timestamp')
	validation_state = Attribute('Validation State', min_version='16.3')
	validity_period = Attribute('Validity Period')
	verisign_ca_additional_field_value = Attribute('VeriSign CA:Additional Field Value')
	verisign_ca_challenge_credential = Attribute('VeriSign CA:Challenge Credential')
	verisign_ca_comment = Attribute('VeriSign CA:Comment')
	verisign_ca_enrollment_mode = Attribute('VeriSign CA:Enrollment Mode')
	verisign_ca_first_pickup_request = Attribute('VeriSign CA:First Pickup Request')
	verisign_ca_license_count = Attribute('VeriSign CA:License Count')
	verisign_ca_original_challenge_credential = Attribute('VeriSign CA:Original Challenge Credential')
	verisign_ca_replacement_reason = Attribute('VeriSign CA:Replacement Reason')
	verisign_ca_server_type = Attribute('VeriSign CA:Server Type')
	verisign_ca_specific_end_date = Attribute('VeriSign CA:Specific End Date')
	verisign_ca_timestamp = Attribute('VeriSign CA:Timestamp')
	verizon_ca_additional_field_value = Attribute('Verizon CA:Additional Field Value', min_version='15.4')
	verizon_ca_challenge_credential = Attribute('Verizon CA:Challenge Credential', min_version='15.4')
	verizon_ca_challenge_credential_hint = Attribute('Verizon CA:Challenge Credential Hint', min_version='15.4')
	verizon_ca_enrollment_mode = Attribute('Verizon CA:Enrollment Mode', min_version='15.4')
	verizon_ca_first_pickup_request = Attribute('Verizon CA:First Pickup Request', min_version='15.4')
	verizon_ca_license_count = Attribute('Verizon CA:License Count', min_version='15.4')
	verizon_ca_number_of_servers = Attribute('Verizon CA:Number Of Servers', min_version='15.4')
	verizon_ca_organization_summary = Attribute('Verizon CA:Organization Summary', min_version='15.4')
	verizon_ca_server_name = Attribute('Verizon CA:Server Name', min_version='15.4')
	verizon_ca_server_type = Attribute('Verizon CA:Server Type', min_version='15.4')
	verizon_ca_tech_email = Attribute('Verizon CA:Tech Email', min_version='15.4')
	verizon_ca_tech_firstname = Attribute('Verizon CA:Tech Firstname', min_version='15.4')
	verizon_ca_tech_surname = Attribute('Verizon CA:Tech Surname', min_version='15.4')
	verizon_ca_tech_telnumber = Attribute('Verizon CA:Tech Telnumber', min_version='15.4')
	verizon_ca_timestamp = Attribute('Verizon CA:Timestamp', min_version='15.4')
	want_renewal = Attribute('Want Renewal')
	x509_d = Attribute('X509 D', min_version='15.2')
	x509_dc = Attribute('X509 DC', min_version='15.2')
	x509_dnq = Attribute('X509 DNQ', min_version='15.2')
	x509_e = Attribute('X509 E', min_version='15.2')
	x509_extension_fields = Attribute('X509 Extension Fields')
	x509_gn = Attribute('X509 GN', min_version='15.2')
	x509_gq = Attribute('X509 GQ', min_version='15.2')
	x509_i = Attribute('X509 I', min_version='15.2')
	x509_p = Attribute('X509 P', min_version='15.2')
	x509_pa = Attribute('X509 PA', min_version='15.2')
	x509_pc = Attribute('X509 PC', min_version='15.2')
	x509_sa = Attribute('X509 SA', min_version='15.2')
	x509_sn = Attribute('X509 SN', min_version='15.2')
	x509_sno = Attribute('X509 SNO', min_version='15.2')
	x509_subject = Attribute('X509 Subject')
	x509_subjectaltname = Attribute('X509 SubjectAltName')
	x509_subjectaltname_dns = Attribute('X509 SubjectAltName DNS')
	x509_subjectaltname_ipaddress = Attribute('X509 SubjectAltName IPAddress')
	x509_subjectaltname_othername_upn = Attribute('X509 SubjectAltName OtherName UPN')
	x509_subjectaltname_rfc822 = Attribute('X509 SubjectAltName RFC822')
	x509_subjectaltname_uri = Attribute('X509 SubjectAltName URI')
	x509_t = Attribute('X509 T', min_version='15.2')
	x509_tn = Attribute('X509 TN', min_version='15.2')
	x509_ua = Attribute('X509 UA', min_version='15.2')
	x509_uid = Attribute('X509 UID', min_version='15.2')
	x509_un = Attribute('X509 UN', min_version='15.2')
	xolphin_ca_address = Attribute('Xolphin CA:Address')
	xolphin_ca_approver_email_address = Attribute('Xolphin CA:Approver Email Address')
	xolphin_ca_approver_first_name = Attribute('Xolphin CA:Approver First Name')
	xolphin_ca_approver_last_name = Attribute('Xolphin CA:Approver Last Name')
	xolphin_ca_approver_phone_number = Attribute('Xolphin CA:Approver Phone Number')
	xolphin_ca_approver_type = Attribute('Xolphin CA:Approver Type')
	xolphin_ca_city = Attribute('Xolphin CA:City')
	xolphin_ca_company = Attribute('Xolphin CA:Company')
	xolphin_ca_country_code = Attribute('Xolphin CA:Country Code')
	xolphin_ca_department = Attribute('Xolphin CA:Department')
	xolphin_ca_effective_polling_interval = Attribute('Xolphin CA:Effective Polling Interval')
	xolphin_ca_kvk_number = Attribute('Xolphin CA:KvK Number')
	xolphin_ca_postbox = Attribute('Xolphin CA:Postbox')
	xolphin_ca_reference_number = Attribute('Xolphin CA:Reference Number')
	xolphin_ca_zip_code = Attribute('Xolphin CA:Zip Code')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.discovery_statistics import DiscoveryStatisticsAttributes
from pytpp.attributes.schedule_base import ScheduleBaseAttributes
from pytpp.attributes.top import TopAttributes


class AgentDiscoveryBaseAttributes(DiscoveryStatisticsAttributes, ScheduleBaseAttributes, TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Agent Discovery Base"
	configuration = Attribute('Configuration')
	discovery_exclusion_dn = Attribute('Discovery Exclusion DN')
	protection_key = Attribute('Protection Key')
	report_dn = Attribute('Report DN')
	status = Attribute('Status')



from pytpp.properties.response_objects.dataclasses import system_status
from pytpp.tools.helpers.date_converter import from_date_string


class SystemStatus:
    @staticmethod
    def Engine(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return system_status.Engine(
            dn=response_object.get('DN'),
            display_name=response_object.get('DisplayName'),
            guid=response_object.get('Guid'),
            id=response_object.get('Id'),
            name=response_object.get('Name'),
        )

    @staticmethod
    def Services(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}

        return system_status.Services(
            vplatform=SystemStatus.Service(response_object.get('vPlatform')),
            log_server=SystemStatus.Service(response_object.get('logServer')),
            iis=SystemStatus.Service(response_object.get('iis')),
        )

    @staticmethod
    def Service(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}

        return system_status.Service(
            modules=response_object.get('modules'),
            time_since_first_seen=from_date_string(response_object.get('timeSinceFirstSeen'), duration_format=True),
            time_since_last_seen=from_date_string(response_object.get('timeSinceLastSeen'), duration_format=True),
            status=response_object.get('Status'),
        )

    @staticmethod
    def SystemStatus(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return system_status.SystemStatus(
            engine_name=response_object.get('engineName'),
            services=SystemStatus.Services(response_object.get('services')),
            version=response_object.get('version'),
        )

    @staticmethod
    def Task(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}

        return system_status.Task(
            display_name=response_object.get('DisplayName'),
            name=response_object.get('Name'),
            start_time=from_date_string(response_object.get('StartTime')),
            stop_time=from_date_string(response_object.get('StopTime')),
            warning_count=response_object.get('WarningCount'),
        )

    @staticmethod
    def UpgradeInfo(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}

        return system_status.UpgradeInfo(
            id=response_object.get('Id'),
            start_time=from_date_string(response_object.get('StartTime')),
            versions=response_object.get('Versions'),
        )

    @staticmethod
    def UpgradeStatus(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}

        return system_status.UpgradeStatus(
            engine=SystemStatus.Engine(response_object.get('Engine')),
            status=response_object.get('Status'),
            upgrade_start_time=from_date_string(response_object.get('UpgradeStartTime')),
            upgrade_stop_time=from_date_string(response_object.get('UpgradeStopTime')),
            tasks_completed=[SystemStatus.Task(t) for t in response_object.get('TasksCompleted', [])],
            tasks_pending=[SystemStatus.Task(t) for t in response_object.get('TasksPending', [])],
            tasks_running=[SystemStatus.Task(t) for t in response_object.get('TasksRunning', [])],
        )

    @staticmethod
    def UpgradeSummary(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}

        return system_status.UpgradeSummary(
            status=response_object.get('Status'),
            upgrade_start_time=from_date_string(response_object.get('UpgradeStartTime')),
            upgrade_stop_time=from_date_string(response_object.get('UpgradeStopTime')),
            completed_tasks=response_object.get('CompletedTasks'),
            target_version=response_object.get('TargetVersion'),
            engines_complete=response_object.get('EnginesComplete'),
            engines_running=response_object.get('EnginesRunning'),
            engines_blocked=response_object.get('EnginesBlocked'),
            engines_in_error=response_object.get('EnginesInError'),
            engines_pending_install=response_object.get('EnginesPendingInstall'),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.log_channel import LogChannelAttributes


class LogSMTPAttributes(LogChannelAttributes, metaclass=IterableMeta):
	__config_class__ = "Log SMTP"
	attachment_behavior = Attribute('Attachment Behavior', min_version='21.2')
	cc = Attribute('CC')
	host = Attribute('Host')
	item_vault_id = Attribute('Item Vault Id')
	log_delivery = Attribute('Log Delivery')
	message_body = Attribute('Message Body')
	recipient = Attribute('Recipient')
	smtp_credentials = Attribute('SMTP Credentials')
	secure = Attribute('Secure')
	sender = Attribute('Sender')
	subject = Attribute('Subject')
	template_vault_id = Attribute('Template Vault Id')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class LogApplicationCustomizationAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Log Application Customization"
	customization_data = Attribute('Customization Data')
	language = Attribute('Language')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.report_base import ReportBaseAttributes


class ReportLicensingAttributes(ReportBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Report:Licensing"
	options = Attribute('Options')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class EntrustPKIGatewayAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Entrust PKI Gateway"
	ca_capabilities = Attribute('CA Capabilities', min_version='20.2')
	ca_name = Attribute('CA Name', min_version='20.2')
	exclude_subjectvariables = Attribute('Exclude SubjectVariables', min_version='20.4')
	profile_capabilities = Attribute('Profile Capabilities', min_version='20.2')
	web_service_url = Attribute('Web Service URL', min_version='20.2')



from dataclasses import dataclass
from datetime import datetime
from typing import List


@dataclass
class Result:
    code: int
    workflow_result: str


@dataclass
class Details:
    approval_explanation: str
    approval_from: str
    approvers: List[str]
    blocking: str
    created: datetime
    issued_due_to: str
    status: str
    updated: datetime



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_trust_store_base import CertificateTrustStoreBaseAttributes


class PKCS12TrustStoreAttributes(CertificateTrustStoreBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "PKCS#12 Trust Store"
	key_store = Attribute('Key Store', min_version='15.1')
	key_store_credential = Attribute('Key Store Credential', min_version='15.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_portal_base import ClientPortalBaseAttributes
from pytpp.attributes.client_work_base import ClientWorkBaseAttributes
from pytpp.attributes.x509_certificate_base import X509CertificateBaseAttributes


class ClientUserCertificateWorkAttributes(ClientPortalBaseAttributes, ClientWorkBaseAttributes, X509CertificateBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Client User Certificate Work"
	certificate_bundle_capacity = Attribute('Certificate Bundle Capacity', min_version='18.3')
	certificate_bundle_capacity_mobile_config = Attribute('Certificate Bundle Capacity Mobile Config', min_version='18.3')
	certificate_container = Attribute('Certificate Container', min_version='15.2')
	certificate_icon = Attribute('Certificate Icon', min_version='18.4')
	download_instructions = Attribute('Download Instructions', min_version='15.4')
	download_limit = Attribute('Download Limit', min_version='15.4')
	include_historic_certificates = Attribute('Include Historic Certificates', min_version='15.2')
	membership_loss_disable = Attribute('Membership Loss Disable', min_version='15.2')
	membership_loss_revoke = Attribute('Membership Loss Revoke', min_version='15.2')
	naming_pattern = Attribute('Naming Pattern', min_version='15.2')
	outlook_profile_generation = Attribute('Outlook Profile Generation', min_version='15.2')
	outlook_profile_name = Attribute('Outlook Profile Name', min_version='15.2')
	outlook_profile_options = Attribute('Outlook Profile Options', min_version='15.2')
	portal_friendly_name = Attribute('Portal Friendly Name', min_version='15.4')
	publish_to_identity = Attribute('Publish To Identity', min_version='15.2')
	publish_to_identity_on_pre_enroll = Attribute('Publish To Identity on Pre-Enroll', min_version='15.2')
	required_member_identity = Attribute('Required Member Identity', min_version='15.2')
	transfer_allowed = Attribute('Transfer Allowed', min_version='18.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_authority_base import CertificateAuthorityBaseAttributes


class RSAKeonCAAttributes(CertificateAuthorityBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "RSA Keon CA"
	ca_md5 = Attribute('CA MD5')
	ca_name = Attribute('CA Name')
	certificate_block = Attribute('Certificate Block')
	jurisdiction_id = Attribute('Jurisdiction ID')
	jurisdiction_name = Attribute('Jurisdiction Name')
	supported_validity_periods = Attribute('Supported Validity Periods')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class StatisticsCounterAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Statistics Counter"
	sensitive = Attribute('Sensitive', min_version='19.3')
	statistic_id = Attribute('Statistic Id', min_version='19.3')
	tag_a_name = Attribute('Tag A Name', min_version='19.3')
	tag_b_name = Attribute('Tag B Name', min_version='19.3')
	tag_c_name = Attribute('Tag C Name', min_version='19.3')
	value_description = Attribute('Value Description', min_version='19.3')



from pytpp.attributes._helper import IterableMeta, Attribute


class ValidationBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Validation Base"
	file_validation_error = Attribute('File Validation Error', min_version='15.3')
	file_validation_result = Attribute('File Validation Result', min_version='15.3')
	last_validation = Attribute('Last Validation')
	last_validation_result = Attribute('Last Validation Result')
	ssl_listen_host = Attribute('SSL Listen Host')
	ssl_listen_port = Attribute('SSL Listen Port')
	use_specified_host = Attribute('Use Specified Host')
	validation_disabled = Attribute('Validation Disabled')
	validation_errors = Attribute('Validation Errors', min_version='15.3')
	validation_results = Attribute('Validation Results', min_version='15.3')



from pytpp.properties.response_objects.dataclasses import worfklow
from pytpp.properties.resultcodes import ResultCodes
from pytpp.tools.helpers.date_converter import from_date_string


class Workflow:
    @staticmethod
    def Result(code: int):
        return worfklow.Result(
            code=code,
            workflow_result=ResultCodes.Workflow.get(code, 'Unknown'),
        )

    @staticmethod
    def Details(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return worfklow.Details(
            approval_explanation=response_object.get('ApprovalExplanation'),
            approval_from=response_object.get('ApprovalFrom'),
            approvers=response_object.get('Approvers'),
            blocking=response_object.get('Blocking'),
            created=from_date_string(response_object.get('Created')),
            issued_due_to=response_object.get('IssuedDueTo'),
            status=response_object.get('Status'),
            updated=from_date_string(response_object.get('Updated')),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class MetadataBaseAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Metadata Base"
	allowed_values = Attribute('Allowed Values')
	associated_classes = Attribute('Associated Classes')
	category = Attribute('Category')
	default_values = Attribute('Default Values')
	help_text = Attribute('Help Text')
	label_text = Attribute('Label Text')
	localization = Attribute('Localization')
	mandatory = Attribute('Mandatory')
	not_before = Attribute('Not Before')
	policyable = Attribute('Policyable')
	render_hidden = Attribute('Render Hidden', min_version='16.2')
	render_read_only = Attribute('Render Read Only', min_version='16.2')



from pytpp.tools.helpers.date_converter import from_date_string
from pytpp.properties.response_objects.dataclasses import certificate


class Certificate:
    @staticmethod
    def Certificate(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}

        return certificate.Certificate(
            created_on=response_object.get('CreatedOn'),
            dn=response_object.get('DN'),
            guid=response_object.get('Guid'),
            name=response_object.get('Name'),
            parent_dn=response_object.get('ParentDn'),
            schema_class=response_object.get('SchemaClass'),
            x509=Certificate._X509(response_object.get('X509')),
            links=[Certificate.Link(link) for link in response_object.get('_links', [])],
        )

    @staticmethod
    def Link(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate.Link(
            details=response_object.get('Details'),
            next=response_object.get('Next'),
            previous=response_object.get('Previous'),
        )

    @staticmethod
    def CSR(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate.CSR(
            details=Certificate._CSRDetails(response_object.get('Details')),
            enrollable=response_object.get('Enrollable'),
        )

    @staticmethod
    def Policy(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate.Policy(
            certificate_authority=Certificate._LockedSingleValue(response_object.get('CertificateAuthority')),
            csr_generation=Certificate._LockedSingleValue(response_object.get('CsrGeneration')),
            key_generation=Certificate._LockedSingleValue(response_object.get('KeyGeneration')),
            key_pair=Certificate._LockedKeyPair(response_object.get('KeyPair')),
            management_type=Certificate._LockedSingleValue(response_object.get('ManagementType')),
            private_key_reuse_allowed=response_object.get('PrivateKeyReuseAllowed'),
            subj_alt_name_dns_allowed=response_object.get('SubjAltNameDnsAllowed'),
            subj_alt_name_email_allowed=response_object.get('SubjAltNameEmailAllowed'),
            subj_alt_name_ip_allowed=response_object.get('SubjAltNameIpAllowed'),
            subj_alt_name_upn_allowed=response_object.get('SubjAltNameUpnAllowed'),
            subj_alt_name_uri_allowed=response_object.get('SubjAltNameUriAllowed'),
            subject=Certificate._LockedSubject(response_object.get('Subject')),
            unique_subject_enforced=response_object.get('UniqueSubjectEnforced'),
            whitelisted_domains=response_object.get('WhitelistedDomains'),
            wildcards_allowed=response_object.get('WildcardsAllowed'),
        )

    @staticmethod
    def CertificateDetails(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate.CertificateDetails(
            c=response_object.get('C'),
            cn=response_object.get('CN'),
            enhanced_key_usage=response_object.get('EnhancedKeyUsage'),
            issuer=response_object.get('Issuer'),
            key_algorithm=response_object.get('KeyAlgorithm'),
            key_size=response_object.get('KeySize'),
            key_usage=response_object.get('KeyUsage'),
            l=response_object.get('L'),
            o=response_object.get('O'),
            ou=response_object.get('OU'),
            public_key_hash=response_object.get('PublicKeyHash'),
            s=response_object.get('S'),
            ski_key_identifier=response_object.get('SKIKeyIdentifier'),
            serial=response_object.get('Serial'),
            signature_algorithm=response_object.get('SignatureAlgorithm'),
            signature_algorithm_oid=response_object.get('SignatureAlgorithmOID'),
            store_added=response_object.get('StoreAdded'),
            subject=response_object.get('Subject'),
            subject_alt_name_dns=response_object.get('SubjectAltNameDNS'),
            subject_alt_name_email=response_object.get('SubjectAltNameEmail'),
            subject_alt_name_ip=response_object.get('SubjectAltNameIp'),
            subject_alt_name_upn=response_object.get('SubjectAltNameUpn'),
            subject_alt_name_uri=response_object.get('SubjectAltNameUri'),
            thumbprint=response_object.get('Thumbprint'),
            valid_from=from_date_string(response_object.get('ValidFrom')),
            valid_to=from_date_string(response_object.get('ValidTo')),
        )

    @staticmethod
    def PreviousVersions(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate.PreviousVersions(
            certificate_details=Certificate.CertificateDetails(response_object.get('CertificateDetails')),
            vault_id=response_object.get('VaultId'),
        )

    @staticmethod
    def ProcessingDetails(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate.ProcessingDetails(
            in_error=response_object.get('InError'),
            stage=response_object.get('Stage'),
            status=response_object.get('Status'),
        )

    @staticmethod
    def RenewalDetails(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate.RenewalDetails(
            city=response_object.get('City'),
            country=response_object.get('Country'),
            organization=response_object.get('Organization'),
            organizational_unit=response_object.get('OrganizationUnit'),
            state=response_object.get('State'),
            subject=response_object.get('Subject'),
            subject_alt_name_dns=response_object.get('SubjectAltNameDNS'),
            subject_alt_name_email=response_object.get('SubjectAltNameEmail'),
            subject_alt_name_ip_address=response_object.get('SubjectAltNameIPAddress'),
            subject_alt_name_other_name_upn=response_object.get('SubjectAltNameOtherNameUPN'),
            subject_alt_name_uri=response_object.get('SubjectAltNameURI'),
            valid_from=from_date_string(response_object.get('ValidFrom')),
            valid_to=from_date_string(response_object.get('ValidTo')),
        )

    @staticmethod
    def ValidationDetails(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate.ValidationDetails(
            last_validation_state_update=response_object.get('LastValidationStateUpdate'),
            validation_state=response_object.get('ValidationState'),
        )

    @staticmethod
    def SslTls(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate.SslTls(
            host=response_object.get('Host'),
            ip_address=response_object.get('IpAddress'),
            port=response_object.get('Port'),
            result=Certificate._SslTlsResult(response_object.get('Result')),
            sources=response_object.get('Sources'),
        )

    @staticmethod
    def File(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate.File(
            installation=response_object.get('Installation'),
            performed_on=from_date_string(response_object.get('PerformedOn')),
            result=response_object.get('Result'),
        )

    @staticmethod
    def _SslTlsResult(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate._SslTlsResult(
            chain=Certificate._BitMaskValues(response_object.get('Chain')),
            end_entity=Certificate._BitMaskValues(response_object.get('EndEntity')),
            id=response_object.get('ID'),
            protocols=Certificate._BitMaskValues(response_object.get('Protocols')),
        )

    @staticmethod
    def _BitMaskValues(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate._BitMaskValues(
            bitmask=response_object.get('BitMask'),
            values=response_object.get('Values'),
        )

    @staticmethod
    def _SANS(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate._SANS(
            dns=response_object.get('DNS'),
            ip=response_object.get('IP'),
        )

    @staticmethod
    def _X509(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate._X509(
            cn=response_object.get('CN'),
            issuer=response_object.get('Issuer'),
            key_algorithm=response_object.get('KeyAlgorithm'),
            key_size=response_object.get('KeySize'),
            sans=response_object.get('SANS'),
            serial=response_object.get('Serial'),
            subject=response_object.get('Subject'),
            thumbprint=response_object.get('Thumbprint'),
            valid_from=from_date_string(response_object.get('ValidFrom')),
            valid_to=from_date_string(response_object.get('ValidTo')),
        )

    @staticmethod
    def _CompliantSingleValue(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate._CompliantSingleValue(
            compliant=response_object.get('Compliant'),
            value=response_object.get('Value'),
        )

    @staticmethod
    def _CompliantMultiValue(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate._CompliantMultiValue(
            compliant=response_object.get('Compliant'),
            values=response_object.get('Values'),
        )

    @staticmethod
    def _LockedSingleValue(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate._LockedSingleValue(
            locked=response_object.get('Locked'),
            value=response_object.get('Value'),
        )

    @staticmethod
    def _LockedMultiValue(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate._LockedMultiValue(
            locked=response_object.get('Locked'),
            values=response_object.get('Values'),
        )

    @staticmethod
    def _LockedKeyPair(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate._LockedKeyPair(
            key_algorithm=Certificate._LockedSingleValue(response_object.get('KeyAlgorithm')),
            key_size=Certificate._LockedSingleValue(response_object.get('KeySize')),
        )

    @staticmethod
    def _LockedSubject(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate._LockedSubject(
            city=Certificate._LockedSingleValue(response_object.get('City')),
            country=Certificate._LockedSingleValue(response_object.get('Country')),
            organization=Certificate._LockedSingleValue(response_object.get('Organization')),
            organizational_units=Certificate._LockedMultiValue(response_object.get('OrganizationalUnit')),
            state=Certificate._LockedSingleValue(response_object.get('State')),
        )

    @staticmethod
    def _CSRDetails(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return certificate._CSRDetails(
            city=Certificate._CompliantSingleValue(response_object.get('City')),
            common_name=Certificate._CompliantSingleValue(response_object.get('CommonName')),
            country=Certificate._CompliantSingleValue(response_object.get('Country')),
            key_algorithm=Certificate._CompliantSingleValue(response_object.get('KeyAlgorithm')),
            key_size=Certificate._CompliantSingleValue(response_object.get('KeySize')),
            organization=Certificate._CompliantSingleValue(response_object.get('Organization')),
            organizational_unit=Certificate._CompliantMultiValue(response_object.get('OrganizationalUnit')),
            private_key_reused=Certificate._CompliantSingleValue(response_object.get('PrivateKeyReused')),
            state=Certificate._CompliantSingleValue(response_object.get('State')),
            subj_alt_name_dns=Certificate._CompliantMultiValue(response_object.get('SubjAltNameDns')),
            subj_alt_name_email=Certificate._CompliantMultiValue(response_object.get('SubjAltNameEmail')),
            subj_alt_name_ip=Certificate._CompliantMultiValue(response_object.get('SubjAltNameIp')),
            subj_alt_name_upn=Certificate._CompliantMultiValue(response_object.get('SubjAltNameUpn')),
            subj_alt_name_uri=Certificate._CompliantMultiValue(response_object.get('SubjAltNameUri')),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.certificate_trust_store_base import CertificateTrustStoreBaseAttributes


class DataPowerTrustStoreAttributes(CertificateTrustStoreBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "DataPower Trust Store"
	application_domain = Attribute('Application Domain', min_version='20.3')
	crl_distribution_points_handling = Attribute('CRL Distribution Points Handling', min_version='20.3')
	certificate_validation_mode = Attribute('Certificate Validation Mode', min_version='20.3')
	check_dates = Attribute('Check Dates', min_version='20.3')
	crypto_validation_credential_name = Attribute('Crypto Validation Credential Name', min_version='20.3')
	folder = Attribute('Folder', min_version='20.3')
	maximum_length = Attribute('Maximum Length', min_version='20.3')
	require_crl = Attribute('Require CRL', min_version='20.3')
	use_crl = Attribute('Use CRL', min_version='20.3')



from pytpp.attributes._helper import IterableMeta, Attribute


class AgentCertificateBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Agent Certificate Base"
	certificate_scanner_capi = Attribute('Certificate Scanner CAPI')
	certificate_scanner_map = Attribute('Certificate Scanner Map')
	certificate_scanner_path = Attribute('Certificate Scanner Path')
	credential = Attribute('Credential')
	digest = Attribute('Digest')
	exclude_remote_mount_points = Attribute('Exclude Remote Mount Points')
	expiration = Attribute('Expiration')
	key_store_validation_disabled = Attribute('Key Store Validation Disabled')
	log_threshold = Attribute('Log Threshold')
	max_filesize = Attribute('Max Filesize')
	maximum_threads = Attribute('Maximum Threads')



from datetime import datetime
from dataclasses import dataclass


@dataclass
class LogEvent:
    client_timestamp: datetime
    component: str
    component_id: int
    component_subsystem: str
    data: str
    grouping: int
    id: int
    name: str
    server_timestamp: datetime
    severity: str
    source_ip: str
    text1: str
    text2: str
    value1: int
    value2: int


@dataclass
class LogEventApplicationDefinition:
    application_name: str
    id: int


@dataclass
class LogEventDefinition:
    data_format: str
    data_title: str
    description: str
    grouping_title: str
    grouping_type: str
    id: int
    text1_title: str
    text2_title: str
    value1_title: str
    value1_type: str
    value2_title: str
    value2_type: str



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class ApplicationGroupAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Application Group"
	certificate = Attribute('Certificate', min_version='19.4')
	common_data_location = Attribute('Common Data Location', min_version='19.4')
	common_data_vault_id = Attribute('Common Data Vault Id', min_version='19.4')
	enrollment_application_dn = Attribute('Enrollment Application DN', min_version='19.4')
	primary_application_dn = Attribute('Primary Application DN', min_version='19.4')
	private_key_stub_vault_id = Attribute('Private Key Stub Vault Id', min_version='19.4')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.branch_base import BranchBaseAttributes


class IntermediateandRootCertificatesAttributes(BranchBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Intermediate and Root Certificates"
	cdp_aia_verification_enabled_on_creation = Attribute('CDP AIA Verification Enabled On Creation')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.password_credential import PasswordCredentialAttributes


class CyberArkPasswordCredentialAttributes(PasswordCredentialAttributes, metaclass=IterableMeta):
	__config_class__ = "CyberArk Password Credential"
	account_name = Attribute('Account Name', min_version='19.3')
	application_id = Attribute('Application ID', min_version='19.3')
	folder = Attribute('Folder', min_version='19.3')
	safe = Attribute('Safe', min_version='19.3')



from pytpp.properties.response_objects.dataclasses import ssh_certificates
from pytpp.properties.resultcodes import ResultCodes


class SSHCertificate:
    @staticmethod
    def Response(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}

        error_code = response_object.get('ErrorCode')
        return ssh_certificates.Response(
            success=response_object.get('Success'),
            error_code=error_code,
            error_message=ResultCodes.SSHErrorCodes.get(error_code, 'Unknown') if error_code else None,
        )

    @staticmethod
    def ProcessingDetails(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return ssh_certificates.ProcessingDetails(
            status=response_object.get('Status'),
            status_description=response_object.get('StatusDescription'),
        )

    @staticmethod
    def CertificateDetails(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return ssh_certificates.CertificateDetails(
            ca_fingerprint_sha256=response_object.get('CAFingerprintSHA256'),
            certificate_fingerprint_sha256=response_object.get('CertificateFingerprintSHA256'),
            certificate_type=response_object.get('CertificateType'),
            extensions=response_object.get('Extensions'),
            key_id=response_object.get('KeyID'),
            key_type=response_object.get('KeyType'),
            principals=response_object.get('Principals'),
            public_key_fingerprint_sha256=response_object.get('PublicKeyFingerprintSHA256'),
            serial_number=response_object.get('SerialNumber'),
            valid_from=response_object.get('ValidFrom'),
            valid_to=response_object.get('ValidTo'),
        )

    @staticmethod
    def RequestDetails(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return ssh_certificates.RequestDetails(
            originating_ip=response_object.get('OriginatingIP'),
            requested_by=response_object.get('RequestedBy'),
        )

    @staticmethod
    def APIClient(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return ssh_certificates.APIClient(
            allowed_to_request_certificate_identifier=response_object.get('AllowedToRequestCertificateIdentifier'),
            allowed_to_request_extensions=response_object.get('AllowedToRequestExtensions'),
            allowed_to_request_force_command=response_object.get('AllowedToRequestForceCommand'),
            allowed_to_request_principals=response_object.get('AllowedToRequestPrincipals'),
            allowed_to_request_source_addresses=response_object.get('AllowedToRequestSourceAddresses'),
        )

    @staticmethod
    def Certificate(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return ssh_certificates.Certificate(
            allowed_private_key_algorithms=response_object.get('AllowedPrivateKeyAlgorithms'),
            allowed_private_key_reuse=response_object.get('AllowedPrivateKeyReuse'),
            certificate_destination_dn=response_object.get('CertificateDestinationDn'),
            default_private_key_algorithm=response_object.get('DefaultPrivateKeyAlgorithm'),
            signature_hashing_algorithm=response_object.get('SignatureHashingAlgorithm'),
            type=response_object.get('Type'),
            validity_period=response_object.get('ValidityPeriod'),
        )

    @staticmethod
    def CAKeyPair(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return ssh_certificates.CAKeyPair(
            created_on=response_object.get('CreatedOn'),
            dn=response_object.get('DN'),
            fingerprint_sha256=response_object.get('FingerprintSHA256'),
            guid=response_object.get('Guid'),
            key_algorithm=response_object.get('KeyAlgorithm'),
            name=response_object.get('Name'),
            public_key_data=response_object.get('PublicKeyData'),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.encryption_driver import EncryptionDriverAttributes


class Pkcs11EncryptionDriverAttributes(EncryptionDriverAttributes, metaclass=IterableMeta):
	__config_class__ = "Pkcs11 Encryption Driver"
	account_type = Attribute('Account Type')
	credential = Attribute('Credential')
	cryptokipath = Attribute('CryptokiPath')
	dsn = Attribute('DSN')
	default_key = Attribute('Default Key', min_version='18.1')
	encrypted_pin = Attribute('Encrypted Pin', min_version='18.1')
	fallback_pin = Attribute('Fallback Pin', min_version='21.2')
	key_generation = Attribute('Key Generation', min_version='18.1')
	key_storage = Attribute('Key Storage', min_version='19.2')
	key_validation = Attribute('Key Validation')
	permitted_keys = Attribute('Permitted Keys')
	slot_id = Attribute('Slot Id')
	token_label = Attribute('Token Label', min_version='20.4')
	token_serial = Attribute('Token Serial', min_version='20.4')
	verigram = Attribute('VeriGram', min_version='18.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class AdaptableAppAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Adaptable App"
	allow_push_without_private_key = Attribute('Allow Push Without Private Key', min_version='20.1')
	certificate_name = Attribute('Certificate Name', min_version='16.3')
	file_validation_disabled = Attribute('File Validation Disabled', min_version='16.3')
	installation_status = Attribute('Installation Status', min_version='16.3')
	log_debug = Attribute('Log Debug', min_version='19.3')
	network_validation_disabled = Attribute('Network Validation Disabled', min_version='16.3')
	option_1 = Attribute('Option 1', min_version='16.3')
	option_2 = Attribute('Option 2', min_version='16.3')
	pbes2_algorithm = Attribute('PBES2 Algorithm', min_version='21.3')
	pk_credential = Attribute('PK Credential', min_version='16.3')
	password_1 = Attribute('Password 1', min_version='16.3')
	powershell_script = Attribute('PowerShell Script', min_version='16.3')
	powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id', min_version='18.4')
	retry_after_script_hash_mismatch = Attribute('Retry After Script Hash Mismatch', min_version='18.4')
	script_execution_timeout = Attribute('Script Execution Timeout', min_version='20.2')
	script_hash_mismatch_error = Attribute('Script Hash Mismatch Error', min_version='18.4')
	secondary_credential = Attribute('Secondary Credential', min_version='17.4')
	text_field_1 = Attribute('Text Field 1', min_version='16.3')
	text_field_2 = Attribute('Text Field 2', min_version='16.3')
	text_field_3 = Attribute('Text Field 3', min_version='16.3')
	text_field_4 = Attribute('Text Field 4', min_version='16.3')
	text_field_5 = Attribute('Text Field 5', min_version='16.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class DataPowerAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "DataPower"
	application_domain = Attribute('Application Domain')
	associate_to_cp = Attribute('Associate To CP')
	certificate_name = Attribute('Certificate Name', min_version='16.1')
	certificate_only = Attribute('Certificate Only')
	chain_cert = Attribute('Chain Cert')
	create_cp = Attribute('Create CP')
	create_ic = Attribute('Create IC')
	create_vc = Attribute('Create VC')
	credential_type = Attribute('Credential Type', min_version='17.3')
	crypto_profile = Attribute('Crypto Profile', min_version='17.3')
	ftp_credential = Attribute('FTP Credential')
	ftp_host = Attribute('FTP Host')
	ftp_path = Attribute('FTP Path')
	ftp_port = Attribute('FTP Port')
	file_validation_disabled = Attribute('File Validation Disabled')
	fips_key = Attribute('Fips Key')
	folder = Attribute('Folder', min_version='17.3')
	max_filesize = Attribute('Max Filesize')
	network_validation_disabled = Attribute('Network Validation Disabled')
	private_key_name = Attribute('Private Key Name', min_version='16.1')
	ssh_prompt = Attribute('SSH Prompt')
	ssl_profile_type = Attribute('SSL Profile Type', min_version='17.3')
	ssl_proxy_profile = Attribute('SSL Proxy Profile')
	temp_certificate_label = Attribute('Temp Certificate Label')
	use_basic_provisioning = Attribute('Use Basic Provisioning', min_version='16.1')
	xml_port = Attribute('XML Port', min_version='17.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.driver_base import DriverBaseAttributes
from pytpp.attributes.schedule_base import ScheduleBaseAttributes


class ReportBaseAttributes(DriverBaseAttributes, ScheduleBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Report Base"
	csv_vault_id = Attribute('CSV Vault Id')
	configuration = Attribute('Configuration')
	creation_date = Attribute('Creation Date')
	credential = Attribute('Credential')
	delivery_type = Attribute('Delivery Type')
	format = Attribute('Format')
	html_vault_id = Attribute('HTML Vault Id')
	internet_email_address = Attribute('Internet EMail Address')
	last_run = Attribute('Last Run')
	options = Attribute('Options')
	pdf_vault_id = Attribute('PDF Vault Id')
	publishing_host = Attribute('Publishing Host')
	publishing_location = Attribute('Publishing Location')
	rtf_vault_id = Attribute('RTF Vault Id')
	report_template = Attribute('Report Template')
	skip_empty = Attribute('Skip Empty')
	title = Attribute('Title')
	xml_vault_id = Attribute('XML Vault Id')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class LogChannelAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Log Channel"
	configuration = Attribute('Configuration')
	driver_arguments = Attribute('Driver Arguments')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.log_channel import LogChannelAttributes


class LogFileAttributes(LogChannelAttributes, metaclass=IterableMeta):
	__config_class__ = "Log File"
	expiration = Attribute('Expiration')
	language = Attribute('Language')
	log_directory = Attribute('Log Directory')
	max_fileage = Attribute('Max Fileage')
	max_filesize = Attribute('Max Filesize')
	translate = Attribute('Translate')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.organization import OrganizationAttributes


class AgentModuleHandlerContainerAttributes(OrganizationAttributes, metaclass=IterableMeta):
	__config_class__ = "Agent Module Handler Container"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.code_signing_environment_base import CodeSigningEnvironmentBaseAttributes


class CodeSigningDotNetEnvironmentAttributes(CodeSigningEnvironmentBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing DotNet Environment"
	signing_key_dn = Attribute('Signing Key DN', min_version='20.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class ACMEChallengeAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "ACME Challenge"
	acme_challenge_token = Attribute('ACME Challenge Token', min_version='21.2')
	acme_challenge_validated_on = Attribute('ACME Challenge Validated On', min_version='21.2')
	status = Attribute('Status', min_version='21.2')
	type = Attribute('Type', min_version='21.2')



from dataclasses import dataclass


@dataclass
class Permissions:
    name: str
    approve: bool = None
    delete: bool = None
    discover: bool = None
    manage: bool = None
    read: bool = None
    revoke: bool = None

    def to_dict(self):
        return {
            'approve' : self.approve,
            'delete'  : self.delete,
            'discover': self.discover,
            'manage'  : self.manage,
            'read'    : self.read,
            'revoke'  : self.revoke,
        }

    def effective(self):
        # "read" is implied when specifying the access permission, so it must be omitted.
        return [k for k, v in self.to_dict().items() if v and k != 'read']


@dataclass
class Scope:
    def __init__(self):
        self._certificate = Permissions(name='certificate')
        self._ssh = Permissions(name='ssh')
        self._codesign = Permissions(name='codesign')
        self._configuration = Permissions(name='configuration')
        self._restricted = Permissions(name='restricted')
        self._security = Permissions(name='security')
        self._statistics = Permissions(name='statistics')
        self._agent = Permissions(name='agent')

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string()

    @property
    def scopes(self):
        return [
            {
                'name'       : p.name,
                'permissions': p.to_dict()
            }
            for p in self.list()
        ]

    def list(self):
        return [
            self._certificate, self._ssh, self._codesign,
            self._configuration, self._restricted,
            self._security, self._statistics, self._agent
        ]

    def to_string(self):
        scopes = [
            f'{p.name}:{",".join(p.effective())}'
            for p in self.list() if p.effective()
        ]
        return ';'.join(scopes)

    def agent(self, delete: bool = False, read: bool = False):
        self._agent.delete = delete
        self._agent.read = read
        return self

    def certificate(self, approve: bool = False, delete: bool = False, discover: bool = False,
                    manage: bool = False, read: bool = False, revoke: bool = False):
        self._certificate.approve = approve
        self._certificate.delete = delete
        self._certificate.discover = discover
        self._certificate.manage = manage
        self._certificate.read = read
        self._certificate.revoke = revoke
        return self

    def configuration(self, delete: bool = False, manage: bool = False, read: bool = False):
        self._configuration.delete = delete
        self._configuration.manage = manage
        self._configuration.read = read
        return self

    def codesign(self, delete: bool = False, manage: bool = False, read: bool = False):
        self._codesign.delete = delete
        self._codesign.manage = manage
        self._codesign.read = read
        return self

    def restricted(self, delete: bool = False, manage: bool = False, read: bool = False):
        self._restricted.delete = delete
        self._restricted.manage = manage
        self._restricted.read = read
        return self

    def security(self, delete: bool = False, manage: bool = False, read: bool = False):
        self._security.delete = delete
        self._security.manage = manage
        self._security.read = read
        return self

    def ssh(self, approve: bool = False, delete: bool = False, discover: bool = False,
            manage: bool = False, read: bool = False):
        self._ssh.approve = approve
        self._ssh.delete = delete
        self._ssh.discover = discover
        self._ssh.manage = manage
        self._ssh.read = read
        return self

    def statistics(self, read: bool = False):
        self._statistics.read = read
        return self



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.tree_root import TreeRootAttributes


class BranchBaseAttributes(TreeRootAttributes, metaclass=IterableMeta):
	__config_class__ = "Branch Base"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class TrustwaveCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Trustwave CA"
	interval = Attribute('Interval')
	reseller_id = Attribute('Reseller ID')
	retrieval_period = Attribute('Retrieval Period')
	web_service_url = Attribute('Web Service URL')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_group import ApplicationGroupAttributes


class ApacheApplicationGroupAttributes(ApplicationGroupAttributes, metaclass=IterableMeta):
	__config_class__ = "Apache Application Group"
	client_tools_path = Attribute('Client Tools Path', min_version='19.4')
	ocs_identifier = Attribute('OCS Identifier', min_version='19.4')
	partition_password_credential = Attribute('Partition Password Credential', min_version='19.4')
	private_key_label = Attribute('Private Key Label', min_version='19.4')
	private_key_location = Attribute('Private Key Location', min_version='19.4')
	protection_type = Attribute('Protection Type', min_version='19.4')
	softcard_identifier = Attribute('Softcard Identifier', min_version='19.4')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.code_signing_environment_template_base import CodeSigningEnvironmentTemplateBaseAttributes


class CodeSigningCSPEnvironmentTemplateAttributes(CodeSigningEnvironmentTemplateBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing CSP Environment Template"
	encryption_key_algorithm = Attribute('Encryption Key Algorithm', min_version='20.2')
	key_container_dn = Attribute('Key Container DN', min_version='20.2')
	key_storage_location = Attribute('Key Storage Location', min_version='20.2')
	max_uses = Attribute('Max Uses', min_version='20.2')
	signing_key_algorithm = Attribute('Signing Key Algorithm', min_version='20.2')
	validity_period = Attribute('Validity Period', min_version='20.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.log_channel import LogChannelAttributes


class LogFilterAttributes(LogChannelAttributes, metaclass=IterableMeta):
	__config_class__ = "Log Filter"
	filter_ids = Attribute('Filter IDs')
	filter_severity = Attribute('Filter Severity')
	log_channel = Attribute('Log Channel')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.code_signing_environment_base import CodeSigningEnvironmentBaseAttributes


class CodeSigningKeyPairEnvironmentAttributes(CodeSigningEnvironmentBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Code Signing Key Pair Environment"
	key_algorithm = Attribute('Key Algorithm', min_version='21.2')
	key_dn = Attribute('Key DN', min_version='21.2')
	key_storage_location = Attribute('Key Storage Location', min_version='21.2')
	max_uses = Attribute('Max Uses', min_version='21.2')
	validity_period = Attribute('Validity Period', min_version='21.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.key_pair import KeyPairAttributes


class SSHCAKeyPairAttributes(KeyPairAttributes, metaclass=IterableMeta):
	__config_class__ = "SSH CA Key Pair"
	fingerprint_sha256 = Attribute('Fingerprint SHA256', min_version='21.2')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.service_module import ServiceModuleAttributes


class CAImportManagerAttributes(ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "CA Import Manager"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.flow_action_base import FlowActionBaseAttributes


class FlowEventActionAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Flow Event Action"
	condition = Attribute('Condition', min_version='19.2')
	data_content = Attribute('Data Content', min_version='19.2')
	event = Attribute('Event', min_version='19.2')
	severity = Attribute('Severity', min_version='19.2')
	text1_content = Attribute('Text1 Content', min_version='19.2')
	text2_content = Attribute('Text2 Content', min_version='19.2')
	value1_content = Attribute('Value1 Content', min_version='19.2')
	value2_content = Attribute('Value2 Content', min_version='19.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.identity_driver import IdentityDriverAttributes


class LDAPIdentityDriverAttributes(IdentityDriverAttributes, metaclass=IterableMeta):
	__config_class__ = "LDAP Identity Driver"
	ambiguous_name_resolution = Attribute('Ambiguous Name Resolution')
	base_dn = Attribute('Base DN')
	configuration = Attribute('Configuration')
	connection_timeout = Attribute('Connection Timeout')
	container_class_name = Attribute('Container Class Name')
	container_query_expression = Attribute('Container Query Expression')
	dsn = Attribute('DSN')
	friendly_name = Attribute('Friendly Name')
	group_class_name = Attribute('Group Class Name')
	group_query_expression = Attribute('Group Query Expression')
	group_search_root = Attribute('Group Search Root')
	host = Attribute('Host')
	internal_identifier = Attribute('Internal Identifier')
	login_name_resolution = Attribute('Login Name Resolution')
	member_identifier = Attribute('Member Identifier')
	memberof_enabled = Attribute('MemberOf Enabled', min_version='15.2')
	membership_resolution = Attribute('Membership Resolution', min_version='15.2')
	port = Attribute('Port')
	resolve_nested_groups = Attribute('Resolve Nested Groups')
	revocation = Attribute('Revocation')
	search_root = Attribute('Search Root')
	secure = Attribute('Secure')
	size_limit = Attribute('Size Limit')
	timeout = Attribute('Timeout')
	title = Attribute('Title')
	universal_identifier = Attribute('Universal Identifier')
	user_class_name = Attribute('User Class Name')
	user_query_expression = Attribute('User Query Expression')
	vault_id = Attribute('Vault Id', min_version='18.2')
	vendor = Attribute('Vendor')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class QuoVadisCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "QuoVadis CA"
	api_credentials = Attribute('API Credentials')
	account_name = Attribute('Account Name')
	account_organization = Attribute('Account Organization')
	certificate_type = Attribute('Certificate Type', min_version='15.4')
	subscriber_email = Attribute('Subscriber Email')
	ui_credentials = Attribute('UI Credentials')
	web_service_url = Attribute('Web Service URL')
	web_ui_url = Attribute('Web UI URL')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class ComodoCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Comodo CA"
	address = Attribute('Address')
	company_number = Attribute('Company Number')
	domain_control_validation = Attribute('Domain Control Validation')
	domain_control_validation_email = Attribute('Domain Control Validation Email')
	postal_code = Attribute('Postal Code')
	uri = Attribute('URI')
